'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, wx
from taskcoachlib.domain.attribute import font


class MixFontsTest(test.TestCase):
    def setUp(self):
        super(MixFontsTest, self).setUp()
        self.mixFonts = font.FontMixer.mix
        self.font = wx.Font(10, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL,
            wx.FONTWEIGHT_NORMAL)
        self.lightFont = wx.Font(self.font.GetPointSize(), 
            self.font.GetFamily(), self.font.GetStyle(), wx.FONTWEIGHT_LIGHT)
        self.boldFont = wx.Font(self.font.GetPointSize(), self.font.GetFamily(),
            self.font.GetStyle(), wx.FONTWEIGHT_BOLD)
        self.italicFont = wx.Font(self.font.GetPointSize(), 
            self.font.GetFamily(), wx.FONTSTYLE_ITALIC, self.font.GetWeight())
        self.underlinedFont = wx.Font(self.font.GetPointSize(), 
            self.font.GetFamily(), self.font.GetStyle(), self.font.GetWeight(), 
            underline=True)
        self.swissFont = wx.Font(self.font.GetPointSize(), wx.FONTFAMILY_SWISS,
            self.font.GetStyle(), self.font.GetWeight())
        self.decorativeFont = wx.Font(self.font.GetPointSize(), 
            wx.FONTFAMILY_DECORATIVE, self.font.GetStyle(), 
            self.font.GetWeight())
        self.romanFont = wx.Font(self.font.GetPointSize(), 
            wx.FONTFAMILY_ROMAN, self.font.GetStyle(), 
            self.font.GetWeight())
        self.scriptFont = wx.Font(self.font.GetPointSize(), 
            wx.FONTFAMILY_SCRIPT, self.font.GetStyle(), 
            self.font.GetWeight())
        self.modernFont = wx.Font(self.font.GetPointSize(), 
            wx.FONTFAMILY_MODERN, self.font.GetStyle(), 
            self.font.GetWeight())
        self.teletypeFont = wx.Font(self.font.GetPointSize(), 
            wx.FONTFAMILY_TELETYPE, self.font.GetStyle(), 
            self.font.GetWeight())
        
    def testOneFont(self):
        self.assertEqual(self.font, self.mixFonts(self.font))

    def testFontSize(self):
        biggerFont = wx.Font(self.font.GetPointSize() + 2, self.font.GetFamily(),
                             self.font.GetStyle(), self.font.GetWeight())
        expectedFontSize = (biggerFont.GetPointSize() + self.font.GetPointSize()) / 2
        self.assertEqual(expectedFontSize, 
                         self.mixFonts(self.font, biggerFont).GetPointSize())

    def testFontWeight_NormalAndBold(self):
        self.assertEqual(wx.FONTWEIGHT_BOLD, 
                         self.mixFonts(self.font, self.boldFont).GetWeight())

    def testFontWeight_NormalAndLight(self):
        self.assertEqual(wx.FONTWEIGHT_LIGHT, 
                         self.mixFonts(self.font, self.lightFont).GetWeight())

    def testFontWeight_LightAndBold(self): 
        self.assertEqual(wx.FONTWEIGHT_NORMAL, 
                         self.mixFonts(self.boldFont, self.lightFont).GetWeight())

    def testFontWeight_TwoLightAndOneBold(self): 
        self.assertEqual(wx.FONTWEIGHT_LIGHT, 
                         self.mixFonts(self.boldFont, self.lightFont,
                                       self.lightFont).GetWeight())

    def testFontStyle_NormalAndItalic(self):
        self.assertEqual(wx.FONTSTYLE_ITALIC,
                         self.mixFonts(self.font, self.italicFont).GetStyle())
        
    def testFontUnderline_NormalAndUnderlined(self):
        self.failUnless(self.mixFonts(self.font, self.underlinedFont).GetUnderlined())
       
    def testFontFamily_DefaultAndSwiss(self):
        self.assertEqual(wx.FONTFAMILY_SWISS, 
                         self.mixFonts(self.font, self.swissFont).GetFamily()) 
        
    @test.skipOnPlatform('__WXGTK__')
    def testFontFamily_DefaultAndDecorative(self):
        self.assertEqual(wx.FONTFAMILY_DECORATIVE, 
                         self.mixFonts(self.font, 
                                       self.decorativeFont).GetFamily()) 
        
    def testFontFamily_DefaultAndRoman(self):
        self.assertEqual(wx.FONTFAMILY_ROMAN, 
                         self.mixFonts(self.font, self.romanFont,
                                       self.romanFont).GetFamily())
        
    @test.skipOnPlatform('__WXGTK__')
    def testFontFamily_DefaultAndScript(self):
        self.assertEqual(wx.FONTFAMILY_SCRIPT, 
                         self.mixFonts(self.font,
                                       self.scriptFont).GetFamily())
        
    @test.skipOnPlatform('__WXGTK__')
    def testFontFamily_DefaultAndModern(self):
        self.assertEqual(wx.FONTFAMILY_MODERN, 
                         self.mixFonts(self.font, 
                                       self.modernFont).GetFamily())

    def testFontFamily_DefaultAndTeletype(self):
        self.assertEqual(wx.FONTFAMILY_TELETYPE, 
                         self.mixFonts(self.font, self.teletypeFont,
                                       self.teletypeFont).GetFamily())

    def testFontFamily_RomandAndSwissAndTwoTeletype(self):
        self.assertEqual(wx.FONTFAMILY_TELETYPE, 
                         self.mixFonts(self.romanFont, self.teletypeFont,
                                       self.swissFont,
                                       self.teletypeFont).GetFamily())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.domain import base
from taskcoachlib.thirdparty.pubsub import pub
from category import Category


class CategoryFilter(base.Filter):
    def __init__(self, *args, **kwargs):
        self.__categories = kwargs.pop('categories')
        self.__filterOnlyWhenAllCategoriesMatch = \
            kwargs.pop('filterOnlyWhenAllCategoriesMatch', False)
        for eventType in (self.__categories.addItemEventType(),
                          self.__categories.removeItemEventType()):
            patterns.Publisher().registerObserver(self.onCategoryChanged,
                                                  eventType=eventType, 
                                                  eventSource=self.__categories)
        eventTypes = (Category.categorizableAddedEventType(),
                      Category.categorizableRemovedEventType(),
                      Category.filterChangedEventType())
        for eventType in eventTypes:
            patterns.Publisher().registerObserver(self.onCategoryChanged,
                                                  eventType=eventType)
        pub.subscribe(self.onFilterMatchingChanged, 'settings.view.categoryfiltermatchall')
        super(CategoryFilter, self).__init__(*args, **kwargs)

    def detach(self):
        super(CategoryFilter, self).detach()
        self.removeObserver(self.onCategoryChanged)

    def filterItems(self, categorizables):
        filteredCategories = self.__categories.filteredCategories()
        if not filteredCategories:
            return categorizables
        
        if self.__filterOnlyWhenAllCategoriesMatch:
            filteredCategorizables = set(categorizables)
            for category in filteredCategories:
                filteredCategorizables &= self.__categorizablesBelongingToCategory(category)
        else:
            filteredCategorizables = set()
            for category in filteredCategories: 
                filteredCategorizables |= self.__categorizablesBelongingToCategory(category)

        filteredCategorizables &= self.observable()
        return filteredCategorizables

    @staticmethod
    def __categorizablesBelongingToCategory(category):
        categorizables = category.categorizables(recursive=True)
        for categorizable in categorizables.copy():
            categorizables |= set(categorizable.children(recursive=True))           
        return categorizables
        
    def onFilterMatchingChanged(self, value):
        self.__filterOnlyWhenAllCategoriesMatch = value
        self.reset()

    def onCategoryChanged(self, event): # pylint: disable=W0613
        self.reset()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2008 Rob McMullen <rob.mcmullen@gmail.com>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.domain import task, note, category, effort, attachment
import wx


class NeedsSelectionMixin(object):
    ''' Mixin class for UI commands that need at least one selected item. ''' 
    def enabled(self, event):
        return super(NeedsSelectionMixin, self).enabled(event) and \
            self.viewer.curselection()


class NeedsSelectedCategorizableMixin(NeedsSelectionMixin):
    ''' Mixin class for UI commands that need at least one selected 
        categorizable. '''
    def enabled(self, event):
        return super(NeedsSelectedCategorizableMixin, self).enabled(event) and \
            (self.viewer.curselectionIsInstanceOf(task.Task) or \
             self.viewer.curselectionIsInstanceOf(note.Note))


class NeedsOneSelectedItemMixin(object):
    ''' Mixin class for UI commands that need exactly one selected item. '''
    def enabled(self, event):
        return super(NeedsOneSelectedItemMixin, self).enabled(event) and \
            len(self.viewer.curselection()) == 1


class NeedsSelectedCompositeMixin(NeedsSelectionMixin):
    ''' Mixin class for UI commands that need at least one selected composite
        item. '''
    def enabled(self, event):
        return super(NeedsSelectedCompositeMixin, self).enabled(event) and \
            (self.viewer.curselectionIsInstanceOf(task.Task) or \
             self.viewer.curselectionIsInstanceOf(note.Note) or \
             self.viewer.curselectionIsInstanceOf(category.Category))

    
class NeedsOneSelectedCompositeItemMixin(NeedsOneSelectedItemMixin, 
                                         NeedsSelectedCompositeMixin):
    ''' Mixin class for UI commands that need exactly one selected composite
        item. '''
    pass


class NeedsAttachmentViewerMixin(object):
    ''' Mixin class for UI commands that need a viewer that is showing
        attachments. '''
    def enabled(self, event):
        return super(NeedsAttachmentViewerMixin, self).enabled(event) and \
            self.viewer.isShowingAttachments()


class NeedsSelectedTasksMixin(NeedsSelectionMixin):
    ''' Mixin class for UI commands that need one or more selected tasks. '''
    def enabled(self, event):
        return super(NeedsSelectedTasksMixin, self).enabled(event) and \
            self.viewer.curselectionIsInstanceOf(task.Task)


class NeedsSelectedNoteOwnersMixin(NeedsSelectionMixin):
    ''' Mixin class for UI commands that need at least one selected note 
        owner. '''
    def enabled(self, event):
        return super(NeedsSelectedNoteOwnersMixin, self).enabled(event) and \
            (self.viewer.curselectionIsInstanceOf(task.Task) or \
             self.viewer.curselectionIsInstanceOf(category.Category) or \
             self.viewer.curselectionIsInstanceOf(attachment.Attachment))


class NeedsSelectedNoteOwnersMixinWithNotes(NeedsSelectedNoteOwnersMixin):
    ''' Mixin class for UI commands that need at least one selected note owner 
        with notes. ''' 
    def enabled(self, event):
        # pylint: disable=E1101
        return super(NeedsSelectedNoteOwnersMixinWithNotes, self).enabled(event) and \
            any([item.notes() for item in self.viewer.curselection()])
            
            
class NeedsSelectedAttachmentOwnersMixin(NeedsSelectionMixin):
    ''' Mixin class for UI commands that need at least one selected attachment 
        owner. '''
    def enabled(self, event):
        return super(NeedsSelectedAttachmentOwnersMixin, self).enabled(event) and \
            (self.viewer.curselectionIsInstanceOf(task.Task) or \
             self.viewer.curselectionIsInstanceOf(category.Category) or \
             self.viewer.curselectionIsInstanceOf(note.Note))


class NeedsOneSelectedTaskMixin(NeedsSelectedTasksMixin, 
                                NeedsOneSelectedItemMixin):
    ''' Mixin class for UI commands that need at least one selected tasks. '''
    pass


class NeedsSelectionWithAttachmentsMixin(NeedsSelectionMixin):
    ''' Mixin class for UI commands that need at least one selected item with
        one or more attachments. '''
    def enabled(self, event):
        return super(NeedsSelectionWithAttachmentsMixin, self).enabled(event) and \
            any(item.attachments() for item in self.viewer.curselection() if not isinstance(item, effort.Effort))


class NeedsSelectedEffortMixin(NeedsSelectionMixin):
    ''' Mixin class for UI commands that need at least one selected effort. '''
    def enabled(self, event):
        return super(NeedsSelectedEffortMixin, self).enabled(event) and \
            self.viewer.curselectionIsInstanceOf(effort.Effort)


class NeedsSelectedAttachmentsMixin(NeedsAttachmentViewerMixin, 
                                    NeedsSelectionMixin):
    ''' Mixin class for UI commands that need at least one selected 
        attachment. '''
    pass


class NeedsAtLeastOneTaskMixin(object):
    ''' Mixin class for UI commands that need at least one task created. '''
    def enabled(self, event):  # pylint: disable=W0613
        return len(self.taskList) > 0


class NeedsAtLeastOneCategoryMixin(object):
    ''' Mixin class for UI commands that need at least one category created. '''
    def enabled(self, event):  # pylint: disable=W0613
        return len(self.categories) > 0
        
        
class NeedsItemsMixin(object):
    ''' Mixin class for UI commands that need at least one item in their 
       viewer. '''
    def enabled(self, event):  # pylint: disable=W0613
        return self.viewer.size() 


class NeedsTreeViewerMixin(object):
    ''' Mixin class for UI commands that need a tree viewer. '''
    def enabled(self, event):
        return super(NeedsTreeViewerMixin, self).enabled(event) and \
            self.viewer.isTreeViewer()


class NeedsDeletedItemsMixin(object):
    ''' Mixin class for UI commands that need deleted items to be present. '''
    def enabled(self, event):
        return super(NeedsDeletedItemsMixin, self).enabled(event) and \
               self.iocontroller.hasDeletedItems()


class PopupButtonMixin(object):
    ''' Mix this with a UICommand for a toolbar pop-up menu. '''

    def doCommand(self, event):  # pylint: disable=W0613
        try:
            args = [self.__menu]
        except AttributeError:
            self.__menu = self.createPopupMenu()  # pylint: disable=W0201
            args = [self.__menu]
        if self.toolbar:
            args.append(self.menuXY())
        self.mainWindow().PopupMenu(*args)  # pylint: disable=W0142

    def menuXY(self):
        ''' Location to pop up the menu. '''
        return self.mainWindow().ScreenToClient((self.menuX(), self.menuY()))

    def menuX(self):
        buttonWidth = self.toolbar.GetToolSize()[0]
        mouseX = wx.GetMousePosition()[0]
        return mouseX - 0.5 * buttonWidth

    def menuY(self):
        toolbarY = self.toolbar.GetScreenPosition()[1]
        toolbarHeight = self.toolbar.GetSize()[1]
        return toolbarY + toolbarHeight
    
    def createPopupMenu(self):
        raise NotImplementedError  # pragma: no cover




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, os, shutil
from taskcoachlib import gui, config

 
class TemplatesDialogTestCase(test.wxTestCase):
    def setUp(self):
        super(TemplatesDialogTestCase, self).setUp()
        self.settings = config.Settings(load=False)

        # Monkey-patching
        self.path = os.path.join(os.path.split(__file__)[0], 'tmpl')
        self.safelyRemove(self.path)
        os.mkdir(self.path)

        self.settings.pathToTemplatesDir = lambda: self.path

        self.editor = gui.dialog.templates.TemplatesDialog(self.settings, 
                                                           self.frame, title='title')
        
    def tearDown(self):
        super(TemplatesDialogTestCase, self).tearDown()
        self.safelyRemove(self.path)
        
    def safelyRemove(self, path):
        try:
            shutil.rmtree(path)
        except OSError: # pragma: no cover
            pass

    def testTwoDefaultTemplates(self):
        self.assertEqual(0, len(self.editor._templates.tasks())) # pylint: disable=W0212



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns 
from taskcoachlib.i18n import _                   
from taskcoachlib import help
from taskcoachlib.thirdparty.pubsub import pub
from taskcoachlib.thirdparty.pubsub.core import Publisher
from taskcoachlib.domain import task
from . import effort


class MaxDateTimeMixin(object):
    def maxDateTime(self):
        stopTimes = [effort.getStop() for effort in self if effort.getStop() is not None]
        return max(stopTimes) if stopTimes else None


class EffortUICommandNamesMixin(object):
    newItemMenuText = _('&New effort...\tCtrl+E')
    newItemHelpText = help.effortNew
    
                        
class EffortList(patterns.SetDecorator, MaxDateTimeMixin, 
                 EffortUICommandNamesMixin):
    ''' EffortList observes a TaskList and contains all effort records of
        all tasks in the underlying TaskList. '''

    def  __init__(self, *args, **kwargs):
        super(EffortList, self).__init__(*args, **kwargs)
        pub.subscribe(self.onAddEffortToOrRemoveEffortFromTask, 
                      task.Task.effortsChangedEventType())
        
    def extendSelf(self, tasks, event=None):
        ''' This method is called when a task is added to the observed list.
            It overrides ObservableListObserver.extendSelf whose default 
            behaviour is to add the item that is added to the observed 
            list to the observing list (this list) unchanged. But we want to 
            add the efforts of the tasks, rather than the tasks themselves. '''
        effortsToAdd = []
        for task in tasks:
            effortsToAdd.extend(task.efforts())
        super(EffortList, self).extendSelf(effortsToAdd, event)
        for effort in effortsToAdd:
            if effort.getStop() is None:
                pub.sendMessage(effort.trackingChangedEventType(), newValue=True, sender=effort)
        
    def removeItemsFromSelf(self, tasks, event=None):
        ''' This method is called when a task is removed from the observed 
            list. It overrides ObservableListObserver.removeItemsFromSelf 
            whose default behaviour is to remove the item that was removed
            from the observed list from the observing list (this list) 
            unchanged. But we want to remove the efforts of the tasks, rather 
            than the tasks themselves. '''
        effortsToRemove = []
        for task in tasks:
            effortsToRemove.extend(task.efforts())
        for effort in effortsToRemove:
            if effort.getStop() is None:
                pub.sendMessage(effort.trackingChangedEventType(), newValue=False, sender=effort)
        super(EffortList, self).removeItemsFromSelf(effortsToRemove, event)

    def onAddEffortToOrRemoveEffortFromTask(self, newValue, sender):
        if sender not in self.observable():
            return
        newValue, oldValue = newValue
        effortsToAdd = [effort for effort in newValue if not effort in oldValue]
        effortsToRemove = [effort for effort in oldValue if not effort in newValue]
        super(EffortList, self).extendSelf(effortsToAdd)
        super(EffortList, self).removeItemsFromSelf(effortsToRemove)
        for effort in effortsToAdd + effortsToRemove:
            if effort.getStop() is None:
                pub.sendMessage(effort.trackingChangedEventType(),
                                newValue=effort in effortsToAdd,
                                sender=effort)

    def originalLength(self):
        ''' Do not delegate originalLength to the underlying TaskList because
            that would return a number of tasks, and not the number of effort 
            records.'''
        return len(self)
        
    def removeItems(self, efforts):  # pylint: disable=W0221
        ''' We override ObservableListObserver.removeItems because the default
            implementation is to remove the arguments from the original list,
            which in this case would mean removing efforts from a task list.
            Since that wouldn't work we remove the efforts from the tasks by
            hand. '''
        for effort in efforts:
            effort.task().removeEffort(effort)

    def extend(self, efforts):  # pylint: disable=W0221
        ''' We override ObservableListObserver.extend because the default
            implementation is to add the arguments to the original list,
            which in this case would mean adding efforts to a task list.
            Since that wouldn't work we add the efforts to the tasks by
            hand. '''
        for effort in efforts:
            effort.task().addEffort(effort)
    
    @classmethod        
    def sortEventType(class_):
        return 'this event type is not used'


class EffortListTracker(patterns.Observer, Publisher):
    ''' EffortListTracker observes an EffortList and keeps track of
    currently tracked efforts. '''

    def __init__(self, effortList, includeComposites=False):
        '''@param effortList: The effort list to observe.
        @param includeComposites: if False, composite efforts will be
            ignored.'''
        super(EffortListTracker, self).__init__()
        Publisher.__init__(self)

        self.__effortList = effortList
        self.__includeComposites = includeComposites

        # __trackedEfforts is a list and not a set because when an effort is
        # moved from one task to another task we might get the event that the
        # effort is (re)added to the effortList before the event that the effort
        # was removed from the effortList. If we would use a set, the effort
        # would be missing from the set after the removal event.    
        self.__trackedEfforts = self.__filterTrackedEfforts(self.__effortList)

        self.registerObserver(self.onEffortAdded, 
                              eventType=self.__effortList.addItemEventType(),
                              eventSource=self.__effortList)
        self.registerObserver(self.onEffortRemoved, 
                              eventType=self.__effortList.removeItemEventType(),
                              eventSource=self.__effortList)
        pub.subscribe(self.onTrackingChanged, 
                      effort.Effort.trackingChangedEventType())

    def trackedEfforts(self):
        return self.__trackedEfforts

    def onEffortAdded(self, event):
        self.__trackedEfforts.extend(self.__filterTrackedEfforts(event.values()))
        self.sendMessage('effortlisttracker.changed.added', efforts=self.__trackedEfforts)

    def onEffortRemoved(self, event):
        for effort in event.values():
            if effort in self.__trackedEfforts:
                self.__trackedEfforts.remove(effort)
        self.sendMessage('effortlisttracker.changed.removed', efforts=self.__trackedEfforts)
        
    def onTrackingChanged(self, newValue, sender):
        if sender.parent() is None and not self.__includeComposites:
            return
        if sender not in self.__effortList:
            return
        if newValue:
            if sender not in self.__trackedEfforts:
                self.__trackedEfforts.extend([sender])
        else:
            if sender in self.__trackedEfforts:
                self.__trackedEfforts.remove(sender) 
        self.sendMessage('effortlisttracker.changed', efforts=self.__trackedEfforts)

    @staticmethod
    def __filterTrackedEfforts(efforts):
        return [effort for effort in efforts if effort.isBeingTracked()]



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import command
from taskcoachlib.domain import attachment, task, note, category
from CommandTestCase import CommandTestCase


class AddAttachmentTestsMixin(object):
    def addAttachment(self, selectedItems=None):
        self.attachment = attachment.FileAttachment('attachment') # pylint: disable=W0201
        addAttachmentCommand = command.AddAttachmentCommand(self.container,
            selectedItems or [], attachments=[self.attachment])
        addAttachmentCommand.do()

    def testAddOneAttachmentToOneItem(self):
        self.addAttachment([self.item1])
        self.assertDoUndoRedo(lambda: self.assertEqual([self.attachment], 
            self.item1.attachments()), lambda: self.assertEqual([], 
            self.item1.attachments()))
            
    def testAddOneAttachmentToTwoItems(self):
        self.addAttachment([self.item1, self.item2])
        self.assertDoUndoRedo(lambda: self.failUnless([self.attachment] == \
            self.item1.attachments() == self.item2.attachments()), 
            lambda: self.failUnless([] == self.item1.attachments() == \
            self.item2.attachments()))


class AddAttachmentTestCase(CommandTestCase):
    ItemClass = ContainerClass = lambda subject: 'Subclass responsibility'
    
    def setUp(self):
        super(AddAttachmentTestCase, self).setUp()
        self.item1 = self.ItemClass(subject='item1')
        self.item2 = self.ItemClass(subject='item2')
        self.container = self.ContainerClass([self.item1, self.item2])


class AddAttachmentCommandWithTasksTest(AddAttachmentTestCase, AddAttachmentTestsMixin):
    ItemClass = task.Task
    ContainerClass = task.TaskList


class AddAttachmentCommandWithNotesTest(AddAttachmentTestCase, AddAttachmentTestsMixin):
    ItemClass = note.Note
    ContainerClass = note.NoteContainer


class AddAttachmentCommandWithCategoriesTest(AddAttachmentTestCase, AddAttachmentTestsMixin):
    ItemClass = category.Category
    ContainerClass = category.CategoryList
    
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from wxReportScheduler import *
from wxSchedule import *
from wxScheduleUtils import *
from wxScheduler import *
from wxSchedulerConstants import *
from wxDrawer import *



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import meta
from taskcoachlib.i18n import _
from tips import showTips
from uicommand import *


_MSURL = "https://www.microsoft.com/en-us/download/details.aspx?id=5638"


def sequence(*text):
    return ''.join(text)


def a_href(text, name):
    return '<a href="#%s">%s</a>' % (name, text)


def a_name(text, name):
    return '<a name="%s">%s</a>' % (name, text)


def h(level, text):
    return '<h%d>%s</h%d>' % (level, text, level)


def h3(text):
    return h(3, text)


def h4(text):
    return h(4, text)


def h5(text):
    return h(5, text)

    
def p(*text):
    return '<p>%s</p>' % '\n'.join(text)


def ul(*li):  # pylint: disable=W0621
    return '<ul>%s</ul>' % '\n'.join(li)


def li(*text):
    return '<li>%s</li>' % '\n'.join(text)


def table(*tr):  # pylint: disable=W0621
    return '<table border=1>%s</table>' % '\n'.join(tr)


def tr(*td):
    td = '\n'.join(['<td>%s</td>' % td for td in td]) 
    return '<tr>%s</tr>\n' % td


_TOC = sequence(
    h3(_('Table of contents')),
    ul(
        li(
            a_href(_('Tasks'), 'tasks'),
            ul(
                li(a_href(_('About tasks'), 'abouttasks')),
                li(a_href(_('Task properties'), 'taskproperties')),
                li(a_href(_('Task states'), 'taskstates')),
                li(a_href(_('Task colors'), 'taskcolor')),
                li(a_href(_('Reminders'), 'reminders')))),
        li(
            a_href(_('Effort'), 'effort'),
            ul(
               li(a_href(_('About effort'), 'abouteffort')),
               li(a_href(_('Effort properties'), 'effortproperties')))),
        li(
            a_href(_('Categories'), 'categories'),
            ul(
                li(a_href(_('About categories'), 'aboutcategories')),
                li(a_href(_('Category properties'), 'categoryproperties')))),
        li(
            a_href(_('Notes'), 'notes'),
            ul(
                li(a_href(_('About notes'), 'aboutnotes')),
                li(a_href(_('Note properties'), 'noteproperties')))),
        li(
            a_href(_('Printing and exporting'), 'printingandexporting'),
            ul(
                li(a_href(_('About printing and exporting'), 
                          'aboutprintingandexporting')),
                li(a_href(_('Printing'), 'printing')),
                li(a_href(_('Exporting'), 'exporting')))),
        li(
            a_href(_('Multi-user usage'), 'multiuser'),
            ul(
                li(a_href(_('About multi-user'), 'aboutmultiuser')),
                li(a_href(_('Storage options'), 'storage')))),
        li(
            a_href(_('E-mailing tasks'), 'emailtasks'),
            ul(
                li(a_href(_('Custom attributes for e-mailing'), 'emailcustom')))),
        li(
            a_href(_('E-mail integration'), 'email'),
            ul(
                li(a_href(_('About e-mail integration'), 'aboutemail')),
                li(a_href(_('Attaching an e-mail to a task'), 'emailattach')),
                li(a_href(_('Creating a task from an e-mail'), 'emailcreate')))),
        li(
            a_href(_('SyncML support'), 'syncml'),
            ul(
                li(a_href(_('What is SyncML'), 'aboutsyncml')),
                li(a_href(_('Setup'), 'syncmlsetup')),
                li(a_href(_('Limitations'), 'syncmllimits')),
                li(a_href(_('Troubleshooting'), 'syncmltrouble')),
                li(a_href(_('Purging deleted items'), 'syncmlpurge')))),
        li(
            a_href(_('iPhone and iPod Touch'), 'iphone'),
            ul(
                li(a_href(_('%(name)s on the iPhone') % meta.metaDict, 'taskcoachiphone')),
                li(a_href(_('Configuration'), 'iphoneconf')),
                li(a_href(_('Troubleshooting'), 'iphonetrouble')))),
        li(
            a_href(_('Android'), 'android'),
            ul(
                li(a_href(_('%(name)s on Android?') % meta.metaDict, 'taskcoachandroid')),
                li(a_href(_('Todo.txt and Todo.txt Touch'), 'todotxt')),
                li(a_href(_('Importing todo.txt'), 'todotxtimporting')),
                li(a_href(_('Exporting todo.txt'), 'todotxtexporting')))),
        li(
            a_href(_('Task templates'), 'templates'),
            ul(
                li(a_href(_('About templates'), 'abouttemplates')),
                li(a_href(_('Using templates'), 'usingtemplates')))),
        li(
            a_href(_('Graphical user interface'), 'gui')),
        li(
            a_href(_('Keyboard shortcuts'), 'shortcuts'))))
  

_taskSection = sequence(
    h3(
        a_name(_('Tasks'), 'tasks')),
    h4(
        a_name(_('About tasks'), 'abouttasks')),
    p(
        _('''Tasks are the basic objects that you manipulate. Tasks can
represent anything from a single little thing you have to do to a complete 
project consisting of different phases and numerous activities.''')),
    h4(
        a_name(_('Task properties'), 'taskproperties')),
    p(
        _('Tasks have the following properties you can change:'),
        ul(
            li(_('Subject: a single line that summarizes the task.')),
            li(_('Description: a multi-line description of the task.')),
            li(_('''Planned start date: the first date on which the task can be started. 
The planned start date defaults to the date the task is created. It can also be 'None' 
indicating that you don't really want to start this task. This can be convenient 
for e.g. registering sick leave.''')),
           li(_('''Due date: the date the task should be finished. 
This can be 'None' indicating that this task has no fixed due date.''')),
           li(_('''Actual start date: the date the task was actually started.
The actual start date can be edited directly, but it is also set when you 
track effort for the task or when you set the percentage completed of a task
to a value between 0% and 100%.''')),
            li(_('''Completion date: this date is 'None' as long as the task has 
not been completed. It is set to the current date when you mark the task as 
completed. The completion date can also be entered manually.''')),
            li(_('''Prerequisites: other tasks that need to be completed before
a task can be started. The task remains inactive until the last prerequisite task is 
completed. Note that if the task has a specific planned start date set, that
date has to be in the past <em>and</em> all prerequisite tasks need to be
completed before the task becomes late.''')),
            li(_('Budget: amount of hours available for the task.')),
            li(_('Hourly fee: the amount of money earned with the task per hour.')),
            li(_('''Fixed fee: the amount of money earned with the task 
regardless of the time spent.''')))),
    p(
        _('The following properties are calculated from the properties above:'),
        ul(
            li(_('Days left: the number of days left until the due date.')),
            li(_('''Dependents: other tasks that can be started when the 
prerequisite task has been completed.''')),
            li(_('Time spent: effort spent on the task.')),
            li(_('Budget left: task budget minus time spent on the task.')),
            li(_('Revenue: hourly fee times hours spent plus fixed fee.')))),
    h4(
        a_name(_('Task states'), 'taskstates')),
    p(
        _('Tasks always have exactly one of the following states:'),
        ul(
            li(_('Active: the actual start date is in the past;')),
            li(_('''Inactive: the task has not been started and/or not all 
prerequisite tasks have been completed;''')),
            li(_('Completed: the task has been completed.')))),
    p(
        _('In addition, tasks can be referenced as:'),
        ul(
            li(_('Overdue: the due date is in the past;')),
            li(_('''Due soon: the due date is soon (what 'soon' is, can be 
changed in the preferences);''')),
            li(_('''Late: the planned start is in the past and the task has 
not been started;''')),
            li(_('Over budget: no budget left;')),
            li(_('Under budget: still budget left;')),
            li(_('No budget: the task has no budget.')))),
    h4(
        a_name(_('Task colors'), 'taskcolors')),
    p(
        _('The text of tasks is colored according to the following rules:'),
        ul(
            li(_('Overdue tasks are red;')),
            li(_('Tasks due soon are orange;')),
            li(_('Active tasks are black text with a blue icon;')),
            li(_('Late tasks are purple;')),
            li(_('Future tasks are gray, and')),
            li(_('Completed tasks are green.'))),   
        _('''This all assumes you have not changed the text colors through the 
preferences dialog, of course.''')),
    p(
        _('''The background color of tasks is determined by the categories the 
task belongs to. See the section about 
<a href="#categoryproperties">category properties</a> below.''')),
    h4(
        a_name(_('Reminders'), 'reminders')),
    p(
        _('''You can set a reminder for a specific date and time. %(name)s will
show a reminder message at that date and time. From the reminder dialog
you can open the task, start tracking effort for the task, or mark the task
completed. It is also possible to snooze the reminder.''') % meta.metaDict),
    p(
        _('''If you have Growl or Snarl installed, you can instruct
%(name)s to use these for reminders in the preferences dialog.''') % meta.metaDict))

_effortSection = sequence(
    h3(
        a_name(_('Effort'), 'effort')),
    h4(
        a_name(_('About effort'), 'abouteffort')),
    p(
        _('''Whenever you spent time on tasks, you can record the amount of time
spent by tracking effort. Select a task and invoke 'Start tracking effort' in
the Effort menu or context menu or via the 'Start tracking effort' toolbar 
button.''')),
    h4(
        a_name(_('Effort properties'), 'effortproperties')),
    p(
        _('Effort records have the following properties you can change:'),
        ul(
            li(_('Task: the task the effort belongs to.')),
            li(_('Start date/time: start date and time of the effort.')),
            li(_('''Stop date/time: stop date and time of the effort. This can be 
'None' as long as you are still working on the task.''')),
            li(_('Description: a multi-line description of the effort.')))),
    p(
        _('The following properties are calculated from the properties above:'),
        ul(
            li(_('Time spent: how much time you have spent working on the task.')),
            li(_('Revenue: money earned with the time spent.')))))   

_categorySection = sequence(
    h3(
        a_name(_('Categories'), 'categories')),
    h4(
        a_name(_('About categories'), 'aboutcategories')),
    p(
        _('''Tasks and notes may belong to one or more categories. First, you 
need to create the category that you want to use via the 'Category' menu. Then, 
you can add items to one or more categories by editing the item and checking the 
relevant categories for that item in the category pane of the edit dialog.''')),
    p(
        _('''You can limit the items shown in the task and notes viewers to one 
or more categories by checking a category in the category viewer. For example, 
if you have a category 'phone calls' and you check that category, the task 
viewers will only show tasks belonging to that category; in other words the 
phone calls you need to make.''')),
    h4(
        a_name(_('Category properties'), 'categoryproperties')),
    p(
        _('Categories have the following properties you can change:'),
        ul(
            li(_('Subject: a single line that summarizes the category.')),
            li(_('Description: a multi-line description of the category.')),
            li(_('''Mutually exclusive subcategories: a check box indicating
whether the subcategories of the category are mutually exclusive. If they are,
items can only belong to one of the subcategories. When filtering, you can only
filter by one of the subcategories at a time.''')),
            li(_('''Appearance properties such as icon, font and colors: 
the appearance properties are used to render the category, but also the items
that belong to that category. If a category has no color, font or icon of its 
own, but it has a parent category with such a property, the parent's property 
will be used. If an item belongs to multiple categories that each have a color 
associated with it, a mixture of those colors will be used to render that 
item.''')))))

_noteSection = sequence(
    h3(
        a_name(_('Notes'), 'notes')),
    h4(
        a_name(_('About notes'), 'aboutnotes')),
    p(
        _('''Notes can be used to capture random information that you want
to keep in your task file. Notes can be stand-alone or be part of other items,
such as tasks and categories. Stand-alone notes are displayed in the notes
viewer. Notes that are part of other items are not displayed in the notes
viewer.''')),
    h4(
        a_name(_('Note properties'), 'noteproperties')),
    p(
        _('Notes have the following properties you can change:'),
        ul(
            li(_('Subject: a single line that summarizes the note.')),
            li(_('Description: a multi-line description of the note.')),
            li(_('Appearance properties such as icon, font and colors.')))))


_printingAndExportingSection = sequence(
    h3(
        a_name(_('Printing and exporting'), 'printingandexporting')),
    h4(
        a_name(_('About printing and exporting'), 'aboutprintingandexporting')),
    p(
        _('''Both printing and exporting work in the same way: when you print
or export data, the data from the active viewer is printed or exported.
Moreover, the data is printed or exported in the same way as the viewer is 
displaying it. The data is printed or exported in the same order as the
viewer is displaying it. The columns that are visible determine what 
details get printed or exported. When you filter items, for example hide
completed tasks, those items don't get printed or exported.''')),
    h4(
        a_name(_('Printing'), 'printing')),
    p(
        _('''Prepare the contents of a viewer, by putting the items in the 
right order, show or hide the appropriate columns and apply the relevant 
filters.''')),
    p(
        _('''You can preview how the print will look
using the File -> Print preview menu item. You can edit the page settings
using File -> Page setup. When printing and the platform supports it, you can 
choose to print all visible items in the active viewer, or just the 
selected items.''')),
    h4(
        a_name(_('Exporting'), 'exporting')),
    p(
        _('''Prepare the contents of a viewer, by putting the items in the 
right order, show or hide the appropriate columns and apply the relevant 
filters.''')),
    p(
        _('''Next, choose the format you want to export to and whether you
want to export all visible items or just the selected ones. Available formats
to export to include CSV (comma separated format), HTML and iCalendar. When
you export to HTML, a CSS file is created that you can edit to change
the appearance of the HTML.''')))

_emailTasksSection = sequence(
    h3(
        a_name(_('E-mailing tasks'), 'emailtasks')),
    h4(
        a_name(_('Custom attributes for e-mailing tasks'), 'emailcustom')),
    p(
        _('''You can alter the behaviour of the e-mail command using custom attributes
in a task description. Those attributes must be on a line by themselves. Supported
attributes are 'cc' and 'to'. Examples:''') % meta.metaDict),
    p(_('[email:to=foo@spam.com]')),
    p(_('[email:cc=bar@spam.com]')),
    )

_multiuserSection = sequence(
    h3(
        a_name(_('Multi-user usage'), 'multiuser')),
    h4(
        a_name(_('About multi-user'), 'aboutmultiuser')),
    p(
        _('''A task file may be opened by several instances of %(name)s, either
running on the same computer or on different ones, on a network share for
instance. When you save, %(name)s will actually merge your work with whatever
has been saved on disk since the last time you did. Conflicts are automatically
resolved, usually by you winning the conflict. This serves two use cases:''') % meta.metaDict),
    ul(
        li(_('''A single user, opening the task file on several computers (work,
home, laptop).''')),
        li(_('''Several users working on the same task file.'''))),
    p(
        _('''The first case is the most common and the most secure. The second
case may be dangerous. Most network disk sharing protocols do not support the
kind of file locking that would make this 100% secure. A list of common protocols
and their behaviour follows.''')),
    h4(
        a_name(_('Storage options'), 'storage')),
    p(
        _('''None of the sharing options discussed here work fully. If two users
save their changes within a few hundreds of milliseconds time frame, data will be lost.''')),
    h5(
        _('SMB/CIFS')),
    p(
        _('''This is the most common protocol: Windows shares and their lookalikes
(Samba). If the server and client don't support certain extensions, Task Coach will not
be able to detect automatically when the file has been modified by someone else.''')),
    h5(
        _('NFS')),
    p(
        _('Not tested yet.')),
    h5(
        _('DropBox')),
    p(
        _('''A popular way to access files from several computers (also see SpiderOak
for a more secure alternative). Changes to the task file are correctly detected by %(name)s
when it's updated.''') % meta.metaDict))


_emailSection = sequence(
    h3(
        a_name(_('E-mail integration'), 'email')),
    h4(
        a_name(_('About e-mail integration'), 'aboutemail')),
    p(
        _('''%(name)s integrates with several mail user
agents, through drag and drop. This has some limitations; e-mails are
copied in a directory next to the %(name)s file, as .eml files and are
later opened using whatever program is associated with this file type
on your system. On the other hand, this allows you to open these
e-mail attachments on a system which is different from the one you
created it first.''')%meta.metaDict),
    p(
        _('''Mail user agents supported include:'''),
        ul(
           li(_('Mozilla Thunderbird')),
           li(_('Microsoft Outlook')),
           li(_('Claws Mail')),
           li(_('Apple Mail')))),
    p(
        _('''Due to a Thunderbird limitation, you can't drag and drop several
e-mails from Thunderbird. This does not apply to Outlook.''')),
    h4(
        a_name(_('Attaching an e-mail to a task'), 'emailattach')),
    p(
        _('''There are two ways to attach an e-mail to a task; you can:'''),
        ul(
            li(_('Drop it on a task either in the task tree or the task list.')),
            li(_('Drop it in the attachment pane in the task editor.')))),
    h4(
        a_name(_('Creating a task from an e-mail'), 'emailcreate')),
    p(
        _('''Dropping an e-mail on an empty part of the task tree or task list
creates a new task. Its subject is the subject of the mail, its
description is its content. Additionally, the mail is automatically
attached to the newly created task.''')))


_syncmlSection = sequence(
    h3(
        a_name(_('SyncML support'), 'syncml')),
    h4(    
        a_name(_('What is SyncML'), 'aboutsyncml')),
    p(
        _('''SyncML is an XML protocol designed to synchronize several
applications with a server. A popular open-source server is <a
href="http://www.funambol.com/" target="_blank">Funambol</a>. Synchronization 
clients are available for many devices and applications (Outlook, Pocket PC,
iPod, iPhone, Evolution, etc...), as well as so-called "connectors"
which allow the server to synchronize with Exchange, Google Calendar,
etc.''')),
    p(
        _('''%(name)s has built-in SyncML client support on Windows and Mac OS X
(provided that you use the supplied binaries). This means you can
setup %(name)s to synchronize with the same SyncML server you
synchronize Outlook with and have all Outlook tasks and notes in
your %(name)s file, as well as %(name)s tasks and notes in Outlook. Or
your Pocket PC.''')%meta.metaDict),
    p(
        _('''On Linux, you must install the SyncML client binding for
Python yourself. A 64 bits Debian package for Python 2.7 is available on
<a target="_blank" href="http://www.taskcoach.org/">taskcoach.org</a>,
at the end of the Linux downloads section.''')),
    h4(
        a_name(_('Setup'), 'syncmlsetup')),
    p(
        _('''This feature is optional and off by default. In order to turn it on,
go to the preferences dialog and check it on the Features page.''')),
    p(
        _('''To setup SyncML, edit the SyncML preferences in Edit/SyncML 
preferences. Fill in the synchronization URL, your ID on the server and choose 
which items to synchronize (tasks and/or notes). The URL depends on the server
you choose; some examples are:'''),
        ul(
            li('''<a href="http://www.horde.org/" target="_blank">The Horde project</a>'''),
            li('''<a href="http://memotoo.com" target="_blank">MemoToo</a>:
http://sync.memotoo.com/syncml''')),
        _('''The database names are pretty standard; the default values 
should work.''')),
    p(
        _('''Each task file has its own client ID, so that two different task 
files will be considered different "devices" by the server.''')),
    h4(
        a_name(_('Limitations'), 'syncmllimits')),
    p(
        _('''Some limitations are due to the fact that, the underlying data 
type being vcalendar, some %(name)s features cannot be presented to the 
server.''')%meta.metaDict,
        ul(
            li(_('Task and category hierarchy are lost to the server.')),
            li(_('Recurrence and reminders are not supported yet.')),
            li(_('Note categories are lost to the server.')),
            li(_('''The conflict detection/resolution system is a workaround 
for a Funambol limitation. It should work in most cases, but if many 
applications synchronize with the same server at the same time, problems may 
arise.''')),
            li(_('Probably some others...')))),
    h4(
        a_name(_('Troubleshooting'), 'syncmltrouble')),
    p(
        _('''The SyncML menu items are only present if your platform is 
supported. Currently supported platforms are:'''),
        ul(
            li(_('Windows, 32 bits (see below)')),
            li(_('Linux, 32 bits')),
            li(_('Mac OS 10.3 and later, both Intel and PPC'))),
        _('''You may experience problems under Windows if you don't have the 
Microsoft Visual 8 runtime installed. You can download it on the
<a target="_blank" href="%s"> Microsoft download site</a>.''')% _MSURL),
    h4(
        a_name(_('Purging deleted items'), 'syncmlpurge')),
    p(
        _('''When SyncML is enabled, deleting a task or a note does not actually
delete it, but rather mark it as deleted. The deleted task or note is actually 
removed from the task or note list on the next synchronization. For this reason, 
if you happen to use the SyncML feature, then disable it without having done a 
sync, there may be some deleted notes or tasks in your task file. This is not a 
problem, but takes a little more disk space.''')),
    p(
        _('''In this case, the "Purge deleted items" menu item in the File menu 
can be used to actually delete these tasks. It is only enabled when you
need it, that is when there are items to purge. Be aware that after doing this,
if you re-enable SyncML and make a sync with the same server you used
previously, all those items will reappear, as the server
doesn't know they're deleted.''')))


_iPhoneSection = sequence(
    h3(
        a_name(_('iPhone, iPod Touch and iPad'), 'iphone')),
    h4(
        a_name(_('%(name)s on the iPhone/iPod Touch/iPad')%meta.metaDict, 'taskcoachiphone')),
    p(
        _('''There is an iPhone/iPod Touch/iPad companion app for %(name)s, 
available on <a 
href="http://itunes.apple.com/WebObjects/MZStore.woa/wa/viewSoftware?id=311403563&mt=8"
target="_blank">Apple's AppStore</a>. (If you don't have iTunes installed on 
your computer, you'll be presented a page where you can download iTunes). 
It supports the following features:''')%meta.metaDict,
        ul(
            li(_('''Basic task attributes: subject, description, dates (with 
recurrence)...''')),
            li(_('Hierarchical tasks and categories')),
            li(_('Time tracking')),
            li(_('Multiple task files')),
            li(_('Two-way synchronization with %(name)s on the desktop')%meta.metaDict)),
    p(
        _('''The application is universal and has a custom iPad UI.''')),
    h4(
        a_name(_('Configuration'), 'iphoneconf')),
    h5(
        _('Configuration on the iPhone/iPod Touch/iPad')),
    p(
        _('''There are some settings for the iPhone/iPod Touch/iPad app in the 
Settings application:'''),
        ul(
            li(_('Show completed: whether to show completed tasks.')),
            li(_('''Show inactive: whether to show inactive tasks (planned start date 
in the future).''')),
            li(_('''Icon position: the LED icon may show up either on the 
left side or the right side of the screen.''')),
            li(_('''Compact mode: if this is enabled, the task list has smaller 
LEDs and doesn't show categories or dates.''')),
            li(_('''Confirm complete: if enabled, a message box will pop up for 
confirmation when you mark a task complete by tapping its LED.''')),
            li(_('''# days due soon: How many days in the future is 
considered "soon".''')))),
    h5(
        _('Configuration on the desktop, all platforms')),
    p(
        _('''Before synchronizing, you must also configure %(name)s on the 
desktop; in the preferences, in the "Features" tab, check "Enable iPhone
synchronization". Restart %(name)s. Now, in the preferences, choose the 
"iPhone" tab and fill in at least the password.''')%meta.metaDict),
    p(
        _('''When you tap the "Sync" button in the category view, %(name)s
will automatically detect running instances of %(name)s on your
desktop and ask you to select one (you may have several instances
running on different computers on your network, or several instances
on the same computer). The name displayed is, by default, some string
identifying the computer it's running on. To customize this, you may
change the "Bonjour service name" in the configuration.''')%meta.metaDict),
    p(
        _('''%(name)s will remember the chosen instance and try it next time
you synchronize; if it's not running, it will prompt you again.''')%meta.metaDict),
    p(
        _('''Note that this synchronization happens through the network; there 
is no need for the device to be connected through USB nor for iTunes to
be running.''')),
    h5(
        _('Configuration on Windows')),
    p(
        _('''On Windows, you must install <a
href="http://support.apple.com/kb/dl999">Bonjour for Windows</a> and
unblock it when asked by the firewall.''')),
    h5(
        _('Configuration on Linux')),
    p(
        _('''On Linux, you must have the <a href="http://avahi.org/">Avahi</a> 
daemon installed and running. Most modern distributions already have it. You 
must also install the dnscompat package; its name depends on your distribution 
(libavahi-compat-libdnssd1 on Ubuntu for instance).''')),
    h4(
        a_name(_('Troubleshooting'), 'iphonetrouble')),
    h5(
        _('''I can't seem to find the iPhone/iPod Touch app on Apple's 
website''')),
    p(
        _('''You need to have iTunes installed on your computer to browse 
Apple's App Store. <a href="http://www.apple.com/itunes/" target="_blank">Get 
iTunes</a>.''')),
    h5(
        _('''My computer doesn't appear in the list when trying to sync''')),
    p(
        _('''Check that your iPhone/iPod Touch is connected to the same network 
your computer is through WiFi.''')),
    h5(
        _("The iPhone can't connect to my computer")),
    p(
        _('If you have a firewall, check that ports 4096-4100 are open.'))))


_androidSection = sequence(
    h3(
        a_name(_('Android'), 'android')),
    h4(
        a_name(_('%(name)s on Android?')%meta.metaDict, 'taskcoachandroid')),
    p(
        _('''No, %(name)s is not available for the Android platform. But,
<a target="_blank" href="http://todotxt.com">Todo.txt Touch</a> is. 
You can export your tasks from %(name)s to the <a target="_blank" 
href="https://github.com/ginatrapani/todo.txt-cli/wiki/The-Todo.txt-Format">Todo.txt 
format</a> and then edit your tasks on your Android device.''')%meta.metaDict),
    h4(
        a_name(_('Todo.txt and Todo.txt Touch'), 'todotxt')),
    p(
        _('''Todo.txt is an open source todo list manager, created by Gina 
Trapani, that works with plain text files. Todo.txt has a command-line 
interface. However, since the file format is plain text, you can also edit your 
tasks with any text editor. Todo.txt Touch is a version of Todo.txt for the 
Android platform. Todo.txt Touch uses a todo.txt file that lives in your
Dropbox folder. %(name)s can import from and export to that todo.txt 
file.''')%meta.metaDict),
    p(
        _('''When exporting to Todo.txt, %(name)s creates another file alongside
the target one, suffixed with "-meta". Don't delete this file; %(name)s uses it
to keep track of external changed to the tasks. When a conflict arises (a task
is modified both by Task Coach and by an external application when importing),
the Todo.txt file wins the conflict.''') % meta.metaDict),
    p(
        _('''Tip: if you save your task file in the todo folder that Todo.txt
Touch creates in your Dropbox folder, and you turn on automatic importing and 
exporting of Todo.txt, %(name)s will keep the task file and the 
todo.txt file in sync. Please note that another file, named after the .txt file
with '-meta' appended, is also generated. This is used internally by %(name)s
in order to keep track of what has been modified in the .txt file; don't modify
or delete it.''')%meta.metaDict),
    h4(
        a_name(_('Importing todo.txt'), 'todotxtimporting')),
    p(
        _('''%(name)s imports task subjects, planned start date, due date, completion 
date, priority, contexts and projects. Contexts and projects are both 
transformed into categories in %(name)s. Projects cannot be transformed into 
parent tasks because Todo.txt allows tasks to belong to multiple projects, 
while %(name)s only allows one parent task per task.''')%meta.metaDict),
    p(
        _('''When importing, %(name)s tries to find matching tasks and 
categories and update those instead of creating new ones. It does the matching
by looking at the subject of the task (or project, or context) and the parent
item if any.''')%meta.metaDict),
    h4(
        a_name(_('Exporting todo.txt'), 'todotxtexporting')),
    p(
        _('''%(name)s exports task subjects, planned start date, due date, completion 
date, priority and categories to the Todo.txt format. Other attributes are not
exported.''')%meta.metaDict),
    p(
        _('''%(name)s supports dates and times, but Todo.txt only supports 
dates, so the time part of start, due, and completion date/times is not
exported.''')%meta.metaDict),
    p(
        _('''The default Todo.txt format only supports planned start dates and 
completion dates, but no due dates. Therefore, %(name)s uses an extension to
export the due date of tasks. Due dates are shown as "due:YYYY-MM-DD" in the
todo.txt file.''')%meta.metaDict),
    p(
        _('''Todo.txt has priorities in the form of a letter ('A'-'Z'). 
%(name)s has numerical priorities. %(name)s priority 1 is transformed into 'A',
2 becomes 'B', etc., and 26 is 'Z'. %(name)s priorities below 1 and above 26
are not exported.''')%meta.metaDict),
    p(
        _('''Categories whose subject starts with a '+' are exported as projects. 
Categories whose subject starts with an '@'-sign are exported as context. 
Categories whose subject doesn't start with '+' or '@' are not exported. 
Subcategories are exported if their parent category is.''')%meta.metaDict))

_templatesSection = sequence(
    h3(
        a_name(_('Task templates'), 'templates')),
    h4(
        a_name(_('About templates'), 'abouttemplates')),
    p(
        _('''Templates are blueprints for new tasks. Right now, the only task 
properties that can be "parameterized" are the dates. When instantiating a 
template, the created task has its dates replaced with dates relative to the 
current date.''')),
    h4(
        a_name(_('Using templates'), 'usingtemplates')),
    p(
        _('''One can create a template by selecting a task (only one) and click 
on the "Save task as template" item in the File menu. All subtasks, notes and 
attachments are part of the template. Only categories are not saved.''')),
    p(
        _('''You can also create a new template from a pre-made template file 
(.tsktmpl); just select "Import template" in the File menu and select the file. 
Template files are stored in a subdirectory of the directory where TaskCoach.ini 
is.''')),
    p(
        _('''In order to instantiate a task template, use the "New task from 
template" menu in the Task menu, or the equivalent toolbar button. When the 
task is created, the due, start and completion dates, if applicable, are 
reevaluated relatively to the current date. That means that if you create a 
template from a task starting today and due tomorrow, every time the template 
is instantiated, the planned start date will be replaced by the current date and the 
due date by the current date plus one day.''')),
    p(
        _('''You can also add templates from the template editor (File/Edit
templates), as well as edit the template's basic properties (dates and
subject). Dates are provided in a human-readable format; the date editor
will become red if %(name)s cannot figure out what it means. Example
dates:''')%meta.metaDict),
    ul(
        li('3 pm tomorrow'),
        li('next saturday'),
        li('noon')
    ),
    p(
        _('''Please note that this system is not localized; you must enter
the dates in english.''')))

_guiSection = sequence(
    h3(
        a_name(_('Graphical user interface'), 'gui')),
    p(
        _('''You can drag and drop viewers to create almost any user interface 
layout you want. When you start dragging a viewer, drop hints will appear to
show where you can drop the viewer. Viewers can also be dropped onto each other
to create notebooks.''')),
    p(
        _('''In the edit dialogs, you can drag and drop tabs to rearrange 
the order or to create a whole different user interface layout by placing tabs 
next to eachother.''')),
    p(
        _('''Subjects and descriptions of tasks, notes and categories can be
edited without opening an edit dialog. Select the item whose subject or
description you want to change and click the item again, either in the subject
column or in the description column. A text control will appear that lets you 
change the subject or description. Hit return to confirm your changes. Hit 
escape to cancel your changes. F2 is a keyboard shortcut for editing the 
subject.'''))
        )
        
_shortcutSection = sequence(
    h3(
        a_name(_('Keyboard shortcuts'), 'shortcuts')),
    p(
        _('''%(name)s has several keyboard shortcuts, listed below. Keyboard 
shortcuts are not configurable at the moment.''')%meta.metaDict),
    p(
        table(
            tr(_('Ctrl-A'), editSelectAll),
            tr(_('Shift-Ctrl-A'), addAttachment),
            tr(_('Ctrl-B'), addNote),
            tr(_('Shift-Ctrl-B'), openAllNotes),
            tr(_('Ctrl-C'), editCopy),
            tr(_('Shift-Ctrl-C'), viewCollapseAll),
            tr(_('Ctrl-D'), taskDecreasePriority),
            tr(_('Shift-Ctrl-D'), taskMinPriority),
            tr(_('Ctrl-E'), effortNew),
            tr(_('Shift-Ctrl-E'), viewExpandAll),
            tr(_('Ctrl-F'), search),
            tr(_('Ctrl-G'), categoryNew),
            tr(_('Ctrl-H'), help),
            tr(_('Ctrl-I'), taskIncreasePriority),
            tr(_('Shift-Ctrl-I'), taskMaxPriority),
            tr(_('Ctrl-J'), noteNew),
            tr(_('Ctrl-M (Linux and Windows)'), mailItem),
            tr(_('Shift-Ctrl-M (Mac OS X)'), mailItem),
            tr(_('Shift-Ctrl-M'), fileMergeDiskChanges),
            tr(_('Ctrl-N (Linux and Mac OS X)'), taskNew),
            tr(_('Shift-Ctrl-N (Linux and Mac OS X)'), _('Insert a new subitem')),
            tr(_('Ctrl-O'), fileOpen),
            tr(_('Shift-Ctrl-O'), openAllAttachments),
            tr(_('Alt-P'), editPreferences),
            tr(_('Ctrl-P'), print_),
            tr(_('Shift-Ctrl-P'), printPageSetup),
            tr(_('Ctrl-Q'), fileQuit),
            tr(_('Ctrl-R'), resetCategoryFilter),
            tr(_('Shift-Ctrl-R'), resetFilter),
            tr(_('Ctrl-S'), fileSave),
            tr(_('Shift-Ctrl-S'), fileSaveAs),
            tr(_('Ctrl-T'), effortStart),
            tr(_('Shift-Ctrl-T'), effortStopOrResume),
            tr(_('Ctrl-V'), editPaste),
            tr(_('Shift-Ctrl-V'), editPasteAsSubitem),
            tr(_('Ctrl-W'), fileClose),
            tr(_('Ctrl-X'), editCut),
            tr(_('Ctrl-Y'), editRedo),
            tr(_('Ctrl-Z'), editUndo),
            tr(_('Enter'), _('Edit the selected item(s) or close a dialog')),
            tr(_('Ctrl-Enter'), _('Mark the selected task(s) (un)completed')),
            tr(_('Escape'), _('Cancel a dialog or move keyboard focus from search control back to viewer')),
            tr(_('Tab'), _('Move keyboard focus to the next field in the dialog')),
            tr(_('Shift-Tab'), _('Move keyboard focus to the previous field in the dialog')),
            tr(_('Ctrl-Tab'), _('Move keyboard focus to the next tab in a notebook control')),
            tr(_('Shift-Ctrl-Tab'), _('Move keyboard focus to the previous tab in a notebook control')),
            tr(_('DELETE'), _('Delete the selected item(s)')),
            tr(_('INSERT (Windows)'), taskNew),
            tr(_('Shift-INSERT (Windows)'), _('Insert a new subitem')),
            tr(_('Ctrl-PgDn'), viewNextViewer),
            tr(_('Ctrl-PgUp'), viewPreviousViewer),
            tr(_('Alt-Down'), _('Pop up menu or drop down box')),
            tr(_('F2'), _('Edit the subject of the selected item in a viewer')))))


helpHTML = _TOC + _taskSection + _effortSection + _categorySection + \
    _noteSection + _printingAndExportingSection + _multiuserSection + _emailTasksSection + _emailSection + \
    _syncmlSection + _iPhoneSection + _androidSection + _templatesSection + \
    _guiSection + _shortcutSection


aboutHTML = _('''<h4>%(name)s - %(description)s</h4>
<h5>Version %(version)s, %(date)s</h5>
<p>By %(author)s &lt;%(author_email)s&gt;<p>
<p><a href="%(url)s" target="_blank">%(url)s</a></p>
<p>%(copyright)s</p>
<p>%(license_notice_html)s</p>
''')%meta.metaDict



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2010 Svetoslav Trochev <sal_electronics@hotmail.com>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.domain import date, categorizable, note, attachment, base
from taskcoachlib.domain.attribute.icon import getImageOpen
from taskcoachlib.thirdparty.pubsub import pub
from taskcoachlib.thirdparty._weakrefset import WeakSet
import status
import weakref
import wx


class Task(note.NoteOwner, attachment.AttachmentOwner,
           categorizable.CategorizableCompositeObject):

    maxDateTime = date.DateTime()
    
    def __init__(self, subject='', description='', 
                 dueDateTime=None, plannedStartDateTime=None, 
                 actualStartDateTime=None, completionDateTime=None,
                 budget=None, priority=0, id=None, hourlyFee=0,  # pylint: disable=W0622
                 fixedFee=0, reminder=None, reminderBeforeSnooze=None, categories=None,
                 efforts=None, shouldMarkCompletedWhenAllChildrenCompleted=None, 
                 recurrence=None, percentageComplete=0, prerequisites=None,
                 dependencies=None, *args, **kwargs):
        kwargs['id'] = id
        kwargs['subject'] = subject
        kwargs['description'] = description
        kwargs['categories'] = categories
        super(Task, self).__init__(*args, **kwargs)
        self.__status = None  # status cache
        self.__dueSoonHours = self.settings.getint('behavior', 'duesoonhours')  # pylint: disable=E1101
        maxDateTime = self.maxDateTime    
        self.__dueDateTime = dueDateTime or maxDateTime
        self.__plannedStartDateTime = plannedStartDateTime or maxDateTime
        self.__actualStartDateTime = actualStartDateTime or maxDateTime
        if completionDateTime is None and percentageComplete == 100:
            completionDateTime = date.Now()
        self.__completionDateTime = completionDateTime or maxDateTime
        percentageComplete = 100 if self.__completionDateTime != maxDateTime else percentageComplete
        self.__percentageComplete = percentageComplete
        self.__budget = budget or date.TimeDelta()
        self._efforts = efforts or []
        self.__priority = priority
        self.__hourlyFee = hourlyFee
        self.__fixedFee = fixedFee
        self.__reminder = reminder or maxDateTime
        self.__reminderBeforeSnooze = reminderBeforeSnooze or self.__reminder
        self.__recurrence = date.Recurrence() if recurrence is None else recurrence
        self.__prerequisites = WeakSet(prerequisites or [])
        self.__dependencies = WeakSet(dependencies or [])
        self.__shouldMarkCompletedWhenAllChildrenCompleted = \
            shouldMarkCompletedWhenAllChildrenCompleted
        for effort in self._efforts:
            effort.setTask(self)
        pub.subscribe(self.__computeRecursiveForegroundColor, 'settings.fgcolor')
        pub.subscribe(self.__computeRecursiveBackgroundColor, 'settings.bgcolor')
        pub.subscribe(self.__computeRecursiveIcon, 'settings.icon')
        pub.subscribe(self.__computeRecursiveSelectedIcon, 'settings.icon')
        pub.subscribe(self.onDueSoonHoursChanged, 'settings.behavior.duesoonhours')
        pub.subscribe(self.onMarkParentCompletedWhenAllChildrenCompletedChanged,
                      'settings.behavior.markparentcompletedwhenallchildrencompleted')

        now = date.Now()
        if now < self.__dueDateTime < maxDateTime:
            date.Scheduler().schedule(self.onOverDue, self.__dueDateTime + date.ONE_SECOND)
            if self.__dueSoonHours:
                dueSoonDateTime = self.__dueDateTime + date.ONE_SECOND - date.TimeDelta(hours=self.__dueSoonHours)
                if dueSoonDateTime > date.Now():
                    date.Scheduler().schedule(self.onDueSoon, dueSoonDateTime)
        if now < self.__plannedStartDateTime < maxDateTime:
            date.Scheduler().schedule(self.onTimeToStart, self.__plannedStartDateTime + date.ONE_SECOND)
            
    @patterns.eventSource
    def __setstate__(self, state, event=None):
        super(Task, self).__setstate__(state, event=event)
        self.setPlannedStartDateTime(state['plannedStartDateTime'])
        self.setActualStartDateTime(state['actualStartDateTime'])
        self.setDueDateTime(state['dueDateTime'])
        self.setCompletionDateTime(state['completionDateTime'])
        self.setPercentageComplete(state['percentageComplete'])
        self.setRecurrence(state['recurrence'])
        self.setReminder(state['reminder'])
        self.setEfforts(state['efforts'])
        self.setBudget(state['budget'])
        self.setPriority(state['priority'])
        self.setHourlyFee(state['hourlyFee'])
        self.setFixedFee(state['fixedFee'])
        self.setPrerequisites(state['prerequisites'])
        self.setDependencies(state['dependencies'])
        self.setShouldMarkCompletedWhenAllChildrenCompleted( \
            state['shouldMarkCompletedWhenAllChildrenCompleted'])
        
    def __getstate__(self):
        state = super(Task, self).__getstate__()
        state.update(dict(dueDateTime=self.__dueDateTime, 
            plannedStartDateTime=self.__plannedStartDateTime,
            actualStartDateTime=self.__actualStartDateTime,
            completionDateTime=self.__completionDateTime,
            percentageComplete=self.__percentageComplete,
            children=self.children(), parent=self.parent(), 
            efforts=self._efforts, budget=self.__budget, 
            priority=self.__priority,
            hourlyFee=self.__hourlyFee, fixedFee=self.__fixedFee, 
            recurrence=self.__recurrence.copy(),
            reminder=self.__reminder,
            prerequisites=set(self.__prerequisites),
            dependencies=set(self.__dependencies),
            shouldMarkCompletedWhenAllChildrenCompleted=self.__shouldMarkCompletedWhenAllChildrenCompleted))
        return state

    def __getcopystate__(self):
        state = super(Task, self).__getcopystate__()
        state.update(dict(plannedStartDateTime=self.__plannedStartDateTime, 
            dueDateTime=self.__dueDateTime, 
            actualStartDateTime=self.__actualStartDateTime, 
            completionDateTime=self.__completionDateTime,
            percentageComplete=self.__percentageComplete,
            efforts=[effort.copy() for effort in self._efforts], 
            budget=self.__budget, priority=self.__priority,
            hourlyFee=self.__hourlyFee, fixedFee=self.__fixedFee, 
            recurrence=self.__recurrence.copy(),
            reminder=self.__reminder, 
            shouldMarkCompletedWhenAllChildrenCompleted=self.__shouldMarkCompletedWhenAllChildrenCompleted))
        return state

    @classmethod
    def monitoredAttributes(class_):
        return categorizable.CategorizableCompositeObject.monitoredAttributes() + \
               ['plannedStartDateTime', 'dueDateTime', 'completionDateTime',
                'percentageComplete', 'recurrence', 'reminder', 'budget',
                'priority', 'hourlyFee', 'fixedFee',
                'shouldMarkCompletedWhenAllChildrenCompleted']

    @patterns.eventSource
    def addCategory(self, *categories, **kwargs):
        if super(Task, self).addCategory(*categories, **kwargs):
            self.recomputeAppearance(True, event=kwargs.pop('event'))

    @patterns.eventSource
    def removeCategory(self, *categories, **kwargs):
        if super(Task, self).removeCategory(*categories, **kwargs):
            self.recomputeAppearance(True, event=kwargs.pop('event'))

    @patterns.eventSource
    def setCategories(self, *categories, **kwargs):
        if super(Task, self).setCategories(*categories, **kwargs):
            self.recomputeAppearance(True, event=kwargs.pop('event'))
                
    def allChildrenCompleted(self):
        ''' Return whether all children (non-recursively) are completed. '''
        children = self.children()
        return all(child.completed() for child in children) if children \
            else False        

    @patterns.eventSource
    def addChild(self, child, event=None):
        if child in self.children():
            return
        wasTracking = self.isBeingTracked(recursive=True)
        super(Task, self).addChild(child, event=event)
        self.childChangeEvent(child, wasTracking, event)
        if self.shouldBeMarkedCompleted():
            self.setCompletionDateTime(child.completionDateTime())
        elif self.completed() and not child.completed():
            self.setCompletionDateTime(self.maxDateTime)
        self.recomputeAppearance(recursive=False, event=event)
        child.recomputeAppearance(recursive=True, event=event)

    @patterns.eventSource
    def removeChild(self, child, event=None):
        if child not in self.children():
            return
        wasTracking = self.isBeingTracked(recursive=True)
        super(Task, self).removeChild(child, event=event)
        self.childChangeEvent(child, wasTracking, event)    
        if self.shouldBeMarkedCompleted(): 
            # The removed child was the last uncompleted child
            self.setCompletionDateTime(date.Now())
        self.recomputeAppearance(recursive=False, event=event)
        child.recomputeAppearance(recursive=True, event=event)
                    
    def childChangeEvent(self, child, wasTracking, event):
        childHasTimeSpent = child.timeSpent(recursive=True)
        childHasBudget = child.budget(recursive=True)
        childHasBudgetLeft = child.budgetLeft(recursive=True)
        childHasRevenue = child.revenue(recursive=True)
        childPriority = child.priority(recursive=True)
        # Determine what changes due to the child being added or removed:
        if childHasTimeSpent:
            self.sendTimeSpentChangedMessage()
        if childHasRevenue:
            self.sendRevenueChangedMessage()
        if childHasBudget:
            self.sendBudgetChangedMessage()
        if childHasBudgetLeft or (childHasTimeSpent and \
                                  (childHasBudget or self.budget())):
            self.sendBudgetLeftChangedMessage()
        if childPriority > self.priority():
            self.sendPriorityChangedMessage()
        isTracking = self.isBeingTracked(recursive=True)
        if wasTracking and not isTracking:
            self.sendTrackingChangedMessage(tracking=False)
        elif not wasTracking and isTracking:
            self.sendTrackingChangedMessage(tracking=True)

    @patterns.eventSource    
    def setSubject(self, subject, event=None):
        super(Task, self).setSubject(subject, event=event)
        # The subject of a dependency of our prerequisites has changed, notify:
        for prerequisite in self.prerequisites():   
            pub.sendMessage(prerequisite.dependenciesChangedEventType(), 
                            newValue=prerequisite.dependencies(), 
                            sender=prerequisite)
        # The subject of a prerequisite of our dependencies has changed, notify:
        for dependency in self.dependencies():
            pub.sendMessage(dependency.prerequisitesChangedEventType(), 
                            newValue=dependency.prerequisites(), 
                            sender=dependency)
    # Due date
            
    def dueDateTime(self, recursive=False):
        if recursive:
            childrenDueDateTimes = [child.dueDateTime(recursive=True) for child in \
                                    self.children() if not child.completed()]
            return min(childrenDueDateTimes + [self.__dueDateTime])
        else:
            return self.__dueDateTime

    def setDueDateTime(self, dueDateTime):
        if dueDateTime == self.__dueDateTime:
            return
        self.__dueDateTime = dueDateTime
        date.Scheduler().unschedule(self.onOverDue)
        date.Scheduler().unschedule(self.onDueSoon)
        if date.Now() <= dueDateTime < self.maxDateTime:
            date.Scheduler().schedule(self.onOverDue, 
                                      dueDateTime + date.ONE_SECOND)
            if self.__dueSoonHours > 0:
                dueSoonDateTime = dueDateTime + date.ONE_SECOND - \
                    date.TimeDelta(hours=self.__dueSoonHours)
                if dueSoonDateTime > date.Now():
                    date.Scheduler().schedule(self.onDueSoon, dueSoonDateTime)
        self.markDirty()
        self.recomputeAppearance()
        pub.sendMessage(self.dueDateTimeChangedEventType(), 
                        newValue=dueDateTime, sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.dueDateTimeChangedEventType(),
                            newValue=dueDateTime, sender=ancestor)

    @classmethod
    def dueDateTimeChangedEventType(class_):
        return 'pubsub.task.dueDateTime'

    def onOverDue(self):
        self.recomputeAppearance()
        
    def onDueSoon(self):
        self.recomputeAppearance()
        
    @staticmethod
    def dueDateTimeSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.dueDateTime(recursive=recursive)
    
    @classmethod
    def dueDateTimeSortEventTypes(class_):
        ''' The event types that influence the due date time sort order. '''
        return (class_.dueDateTimeChangedEventType(),)
    
    # Planned start date
    
    def plannedStartDateTime(self, recursive=False):
        if recursive:
            childrenPlannedStartDateTimes = [child.plannedStartDateTime(recursive=True) for child in \
                                      self.children() if not child.completed()]
            return min(childrenPlannedStartDateTimes + [self.__plannedStartDateTime])
        else:
            return self.__plannedStartDateTime
        
    def setPlannedStartDateTime(self, plannedStartDateTime):
        if plannedStartDateTime == self.__plannedStartDateTime:
            return
        self.__plannedStartDateTime = plannedStartDateTime
        date.Scheduler().unschedule(self.onTimeToStart)
        self.markDirty()
        self.recomputeAppearance()
        if plannedStartDateTime < self.maxDateTime:
            date.Scheduler().schedule(self.onTimeToStart, 
                                      plannedStartDateTime + date.ONE_SECOND)
        pub.sendMessage(self.plannedStartDateTimeChangedEventType(), 
                        newValue=plannedStartDateTime, sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.plannedStartDateTimeChangedEventType(),
                            newValue=plannedStartDateTime, sender=ancestor)

    @classmethod
    def plannedStartDateTimeChangedEventType(class_):
        return 'pubsub.task.plannedStartDateTime'
    
    def onTimeToStart(self):
        self.recomputeAppearance()

    @staticmethod
    def plannedStartDateTimeSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.plannedStartDateTime(recursive=recursive)
    
    @classmethod
    def plannedStartDateTimeSortEventTypes(class_):
        ''' The event types that influence the planned start date time sort 
            order. '''
        return (class_.plannedStartDateTimeChangedEventType(),)

    def timeLeft(self, recursive=False):
        return self.dueDateTime(recursive) - date.Now()

    @staticmethod
    def timeLeftSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.timeLeft(recursive=recursive)
    
    @classmethod
    def timeLeftSortEventTypes(class_):
        ''' The event types that influence the time left sort order. '''
        return (class_.dueDateTimeChangedEventType(),)
    
    # Actual start date
    
    def actualStartDateTime(self, recursive=False):
        if recursive:
            childrenActualStartDateTimes = [child.actualStartDateTime(recursive=True) for child in \
                                      self.children() if not child.completed()]
            return min(childrenActualStartDateTimes + [self.__actualStartDateTime])
        else:
            return self.__actualStartDateTime
    
    def setActualStartDateTime(self, actualStartDateTime, recursive=False):
        if actualStartDateTime == self.__actualStartDateTime:
            return
        self.__actualStartDateTime = actualStartDateTime
        if recursive:
            for child in self.children(recursive=True):
                child.setActualStartDateTime(actualStartDateTime)
        self.markDirty()
        self.recomputeAppearance()
        pub.sendMessage(self.actualStartDateTimeChangedEventType(), 
                        newValue=actualStartDateTime, sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.actualStartDateTimeChangedEventType(),
                            newValue=actualStartDateTime, sender=ancestor)

    @classmethod
    def actualStartDateTimeChangedEventType(class_):
        return 'pubsub.task.actualStartDateTime'

    @staticmethod
    def actualStartDateTimeSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.actualStartDateTime(recursive=recursive)
    
    @classmethod
    def actualStartDateTimeSortEventTypes(class_):
        ''' The event types that influence the actual start date time sort order. '''
        return (class_.actualStartDateTimeChangedEventType(),)
        
    # Completion date
            
    def completionDateTime(self, recursive=False):
        if recursive:
            childrenCompletionDateTimes = [child.completionDateTime(recursive=True) \
                for child in self.children() if child.completed()]
            return max(childrenCompletionDateTimes + [self.__completionDateTime])
        else:
            return self.__completionDateTime

    def setCompletionDateTime(self, completionDateTime=None):
        completionDateTime = completionDateTime or date.Now()
        if completionDateTime == self.__completionDateTime:
            return
        if completionDateTime != self.maxDateTime and self.recurrence():
            self.recur(completionDateTime)
        else:
            parent = self.parent()
            if parent:
                oldParentPriority = parent.priority(recursive=True)
            self.__status = None
            self.__completionDateTime = completionDateTime
            if parent and parent.priority(recursive=True) != oldParentPriority:
                parent.sendPriorityChangedMessage()           
            if completionDateTime != self.maxDateTime:
                self.setReminder(None)
                self.setPercentageComplete(100)
            elif self.percentageComplete() == 100:
                self.setPercentageComplete(0)
            if parent:
                if self.completed():
                    if parent.shouldBeMarkedCompleted():
                        parent.setCompletionDateTime(completionDateTime)
                else:
                    if parent.completed():
                        parent.setCompletionDateTime(self.maxDateTime)
            if self.completed():
                for child in self.children():
                    if not child.completed():
                        child.setRecurrence()
                        child.setCompletionDateTime(completionDateTime)
                if self.isBeingTracked():
                    self.stopTracking()                    
            self.recomputeAppearance()
            for dependency in self.dependencies():
                dependency.recomputeAppearance(recursive=True)
            pub.sendMessage(self.completionDateTimeChangedEventType(), 
                        newValue=completionDateTime, sender=self)
            for ancestor in self.ancestors():
                pub.sendMessage(ancestor.completionDateTimeChangedEventType(),
                                newValue=completionDateTime, sender=ancestor)
            
    @classmethod
    def completionDateTimeChangedEventType(class_):
        return 'pubsub.task.completionDateTime'

    def shouldBeMarkedCompleted(self):
        ''' Return whether this task should be marked completed. It should be
            marked completed when 1) it's not completed, 2) all of its children
            are completed, 3) its setting says it should be completed when
            all of its children are completed. '''
        shouldMarkCompletedAccordingToSetting = \
            self.settings.getboolean('behavior',  # pylint: disable=E1101
                'markparentcompletedwhenallchildrencompleted')
        shouldMarkCompletedAccordingToTask = \
            self.shouldMarkCompletedWhenAllChildrenCompleted()
        return ((shouldMarkCompletedAccordingToTask == True) or \
                ((shouldMarkCompletedAccordingToTask == None) and \
                  shouldMarkCompletedAccordingToSetting)) and \
               (not self.completed()) and self.allChildrenCompleted()
      
    @staticmethod  
    def completionDateTimeSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.completionDateTime(recursive=recursive)

    @classmethod
    def completionDateTimeSortEventTypes(class_):
        ''' The event types that influence the completion date time sort order. '''
        return (class_.completionDateTimeChangedEventType(),)
    
    def onMarkParentCompletedWhenAllChildrenCompletedChanged(self, value):
        ''' When the global setting changes, send a percentage completed 
            changed if necessary. '''
        if self.shouldMarkCompletedWhenAllChildrenCompleted() is None and \
            any([child.percentageComplete(True) for child in self.children()]):
            pub.sendMessage(self.percentageCompleteChangedEventType(),
                            newValue=self.percentageComplete(), sender=self)

    # Task state
    
    def completed(self):
        ''' A task is completed if it has a completion date/time. '''
        return self.status() == status.completed

    def overdue(self):
        ''' A task is over due if its due date/time is in the past and it is
            not completed. Note that an over due task is also either active 
            or inactive. '''
        return self.status() == status.overdue

    def inactive(self):
        ''' A task is inactive if it is not completed and either has no planned 
            start date/time or a planned start date/time in the future, and/or 
            its prerequisites are not completed. '''
        return self.status() == status.inactive
        
    def active(self):
        ''' A task is active if it has a planned start date/time in the past and 
            it is not completed. Note that over due and due soon tasks are also 
            considered to be active. So the statuses active, inactive and 
            completed are disjunct, but the statuses active, due soon and over 
            due are not. '''
        return self.status() == status.active

    def dueSoon(self):
        ''' A task is due soon if it is not completed and there is still time 
            left (i.e. it is not over due). '''
        return self.status() == status.duesoon

    def late(self):
        ''' A task is late if it is not active and its planned start date time
            is in the past. '''
        return self.status() == status.late
    
    @classmethod
    def possibleStatuses(class_):
        return (status.inactive, status.late, status.active,
                status.duesoon, status.overdue, status.completed)

    def status(self):
        if self.__status:
            return self.__status
        if self.completionDateTime() != self.maxDateTime:
            self.__status = status.completed
        else:
            now = date.Now()
            if self.dueDateTime() < now: 
                self.__status = status.overdue
            elif 0 <= self.timeLeft().hours() < self.__dueSoonHours:
                self.__status = status.duesoon
            elif self.actualStartDateTime() <= now:
                self.__status = status.active
            # Don't call prerequisite.completed() because it will lead to infinite
            # recursion in the case of circular dependencies:
            elif any([prerequisite.completionDateTime() == self.maxDateTime \
                      for prerequisite in self.prerequisites(recursive=True, 
                                                             upwards=True)]):
                self.__status = status.inactive
            elif self.plannedStartDateTime() < now:
                self.__status = status.late
            else:
                self.__status = status.inactive
        return self.__status
    
    def onDueSoonHoursChanged(self, value):
        date.Scheduler().unschedule(self.onDueSoon)
        self.__dueSoonHours = value
        dueDateTime = self.dueDateTime()
        if dueDateTime < self.maxDateTime:
            newDueSoonDateTime = dueDateTime + date.ONE_SECOND - date.TimeDelta(hours=self.__dueSoonHours)
            date.Scheduler().schedule(self.onDueSoon, newDueSoonDateTime)
        self.recomputeAppearance()
            
    # effort related methods:

    def efforts(self, recursive=False):
        childEfforts = []
        if recursive:
            for child in self.children():
                childEfforts.extend(child.efforts(recursive=True))
        return self._efforts + childEfforts

    def isBeingTracked(self, recursive=False):
        return self.activeEfforts(recursive)

    def activeEfforts(self, recursive=False):
        return [effort for effort in self.efforts(recursive) \
            if effort.isBeingTracked()]
    
    def addEffort(self, effort):
        if effort in self._efforts:
            return
        wasTracking = self.isBeingTracked()
        oldValue = self._efforts[:]
        self._efforts.append(effort)
        if effort.getStart() < self.actualStartDateTime():
            self.setActualStartDateTime(effort.getStart())
        pub.sendMessage(self.effortsChangedEventType(), newValue=(self._efforts,
                        oldValue), sender=self)
        if effort.isBeingTracked() and not wasTracking:
            self.sendTrackingChangedMessage(tracking=True)
        self.sendTimeSpentChangedMessage()
  
    @classmethod
    def effortsChangedEventType(class_):
        return 'pubsub.task.efforts'
          
    def sendTrackingChangedMessage(self, tracking):
        self.recomputeAppearance()  
        pub.sendMessage(self.trackingChangedEventType(), newValue=tracking,
                        sender=self)  
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.trackingChangedEventType(), 
                            newValue=tracking, sender=ancestor)

    def removeEffort(self, effort):
        if effort not in self._efforts:
            return
        oldValue = self._efforts[:]
        self._efforts.remove(effort)
        pub.sendMessage(self.effortsChangedEventType(), newValue=(self._efforts,
                        oldValue), sender=self)
        if effort.isBeingTracked() and not self.isBeingTracked():
            self.sendTrackingChangedMessage(tracking=False)
        self.sendTimeSpentChangedMessage()

    def stopTracking(self):
        for effort in self.activeEfforts():
            effort.setStop()
  
    def setEfforts(self, efforts):
        if efforts == self._efforts:
            return
        oldValue = self._efforts[:]
        self._efforts = efforts
        pub.sendMessage(self.effortsChangedEventType(), newValue=(self._efforts,
                        oldValue), sender=self)
        self.sendTimeSpentChangedMessage()
        
    @classmethod
    def trackingChangedEventType(class_):
        return 'pubsub.task.track'

    # Time spent
    
    def timeSpent(self, recursive=False):
        return sum((effort.duration() for effort in self.efforts(recursive)), 
                   date.TimeDelta())
        
    def sendTimeSpentChangedMessage(self):
        pub.sendMessage(self.timeSpentChangedEventType(), 
                        newValue=self.timeSpent(), sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.timeSpentChangedEventType(), 
                            newValue=ancestor.timeSpent(),
                            sender=ancestor)
        if self.budget(recursive=True):
            self.sendBudgetLeftChangedMessage()
        if self.hourlyFee() > 0:
            self.sendRevenueChangedMessage()
            
    @classmethod
    def timeSpentChangedEventType(class_):
        return 'pubsub.task.timeSpent'

    @staticmethod
    def timeSpentSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.timeSpent(recursive=recursive)
    
    @classmethod
    def timeSpentSortEventTypes(class_):
        ''' The event types that influence the time spent sort order. '''
        return (class_.timeSpentChangedEventType(),)
    
    # Budget
    
    def budget(self, recursive=False):
        result = self.__budget
        if recursive:
            for task in self.children():
                result += task.budget(recursive)
        return result
    
    def setBudget(self, budget):
        if budget == self.__budget:
            return
        self.__budget = budget
        self.sendBudgetChangedMessage()
        self.sendBudgetLeftChangedMessage()
        
    def sendBudgetChangedMessage(self):
        pub.sendMessage(self.budgetChangedEventType(), newValue=self.budget(),
                        sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.budgetChangedEventType(), 
                            newValue=ancestor.budget(recursive=True),
                            sender=ancestor)
            
    @classmethod
    def budgetChangedEventType(class_):
        return 'pubsub.task.budget'
    
    @staticmethod
    def budgetSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.budget(recursive=recursive)
    
    @classmethod
    def budgetSortEventTypes(class_):
        ''' The event types that influence the budget sort order. '''
        return (class_.budgetChangedEventType(),)

    # Budget left
    
    def budgetLeft(self, recursive=False):
        budget = self.budget(recursive)
        return budget - self.timeSpent(recursive) if budget else budget
    
    def sendBudgetLeftChangedMessage(self):
        pub.sendMessage(self.budgetLeftChangedEventType(), 
                        newValue=self.budgetLeft(), sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.budgetLeftChangedEventType(),
                            newValue=ancestor.budgetLeft(recursive=True),
                            sender=ancestor)
            
    @classmethod
    def budgetLeftChangedEventType(class_):
        return 'pubsub.task.budgetLeft'

    @staticmethod
    def budgetLeftSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.budgetLeft(recursive=recursive)

    @classmethod
    def budgetLeftSortEventTypes(class_):
        ''' The event types that influence the budget left sort order. '''
        return (class_.budgetLeftChangedEventType(),)

    # Foreground color

    def setForegroundColor(self, *args, **kwargs):
        super(Task, self).setForegroundColor(*args, **kwargs)
        self.__computeRecursiveForegroundColor()
    
    def foregroundColor(self, recursive=False):
        if not recursive:
            return super(Task, self).foregroundColor(recursive)
        try:
            return self.__recursiveForegroundColor
        except AttributeError:
            return self.__computeRecursiveForegroundColor()
        
    def __computeRecursiveForegroundColor(self, value=None):  # pylint: disable=W0613
        ownColor = super(Task, self).foregroundColor(False)
        if ownColor:
            recursiveColor = ownColor
        else:
            categoryColor = self._categoryForegroundColor()
            if categoryColor:
                recursiveColor = categoryColor
            else:
                recursiveColor = self.statusFgColor()
        self.__recursiveForegroundColor = recursiveColor  # pylint: disable=W0201
        return recursiveColor
    
    def statusFgColor(self):
        ''' Return the current color of task, based on its status (completed,
            overdue, duesoon, inactive, or active). '''            
        return self.fgColorForStatus(self.status())
    
    @classmethod
    def fgColorForStatus(class_, taskStatus):
        return wx.Colour(*eval(class_.settings.get('fgcolor', '%stasks' % taskStatus)))  # pylint: disable=E1101

    def appearanceChangedEvent(self, event):
        self.__computeRecursiveForegroundColor()
        self.__computeRecursiveBackgroundColor()
        self.__computeRecursiveIcon()
        self.__computeRecursiveSelectedIcon()
        super(Task, self).appearanceChangedEvent(event)
        for eachEffort in self.efforts():
            eachEffort.appearanceChangedEvent(event)        

    # Background color
    
    def setBackgroundColor(self, *args, **kwargs):
        super(Task, self).setBackgroundColor(*args, **kwargs)
        self.__computeRecursiveBackgroundColor()
        
    def backgroundColor(self, recursive=False):
        if not recursive:
            return super(Task, self).backgroundColor(recursive)
        try:
            return self.__recursiveBackgroundColor
        except AttributeError:
            return self.__computeRecursiveBackgroundColor()
        
    def __computeRecursiveBackgroundColor(self, *args, **kwargs):  # pylint: disable=W0613
        ownColor = super(Task, self).backgroundColor(recursive=False)
        if ownColor:
            recursiveColor = ownColor
        else:
            categoryColor = self._categoryBackgroundColor()
            if categoryColor:
                recursiveColor = categoryColor
            else:
                statusColor = self.statusBgColor()
                if statusColor:
                    recursiveColor = statusColor
                elif self.parent():
                    recursiveColor = self.parent().backgroundColor(recursive=True)
                else:
                    recursiveColor = None
        self.__recursiveBackgroundColor = recursiveColor  # pylint: disable=W0201
        return recursiveColor
    
    def statusBgColor(self):
        ''' Return the current color of task, based on its status (completed,
            overdue, duesoon, inactive, or active). '''            
        color = self.bgColorForStatus(self.status())
        return None if color == wx.WHITE else color
    
    @classmethod
    def bgColorForStatus(class_, taskStatus):
        return wx.Colour(*eval(class_.settings.get('bgcolor', '%stasks' % taskStatus)))  # pylint: disable=E1101
    
    # Font

    def font(self, recursive=False):
        ownFont = super(Task, self).font(recursive=False)
        if ownFont or not recursive:
            return ownFont
        else:
            categoryFont = self._categoryFont()
            if categoryFont:
                return categoryFont
            else:
                return self.statusFont()

    def statusFont(self):
        ''' Return the current font of task, based on its status (completed,
            overdue, duesoon, inactive, or active). '''
        return self.fontForStatus(self.status())            

    @classmethod
    def fontForStatus(class_, taskStatus):
        nativeInfoString = class_.settings.get('font', '%stasks' % taskStatus)  # pylint: disable=E1101
        return wx.FontFromNativeInfoString(nativeInfoString) if nativeInfoString else None
                
    # Icon
    
    def icon(self, recursive=False):
        if recursive and self.isBeingTracked():
            return 'clock_icon'
        myIcon = super(Task, self).icon()
        if recursive and not myIcon:
            try:
                myIcon = self.__recursiveIcon
            except AttributeError:
                myIcon = self.__computeRecursiveIcon()
        return self.pluralOrSingularIcon(myIcon, native=super(Task, self).icon() == '')
    
    def __computeRecursiveIcon(self, *args, **kwargs):  # pylint: disable=W0613
        # pylint: disable=W0201
        self.__recursiveIcon = self.categoryIcon() or self.statusIcon()
        return self.__recursiveIcon

    def selectedIcon(self, recursive=False):
        if recursive and self.isBeingTracked():
            return 'clock_icon'
        myIcon = super(Task, self).selectedIcon()
        if recursive and not myIcon:
            try:
                myIcon = self.__recursiveSelectedIcon
            except AttributeError:
                myIcon = self.__computeRecursiveSelectedIcon()
        return self.pluralOrSingularIcon(myIcon, native=super(Task, self).selectedIcon == '')
        
    def __computeRecursiveSelectedIcon(self, *args, **kwargs):  # pylint: disable=W0613
        # pylint: disable=W0201
        self.__recursiveSelectedIcon = self.categorySelectedIcon() or self.statusIcon(selected=True)
        return self.__recursiveSelectedIcon

    def statusIcon(self, selected=False):
        ''' Return the current icon of the task, based on its status. '''
        return self.iconForStatus(self.status(), selected)            

    def iconForStatus(self, taskStatus, selected=False):
        iconName = self.settings.get('icon', '%stasks' % taskStatus)  # pylint: disable=E1101
        iconName = self.pluralOrSingularIcon(iconName)
        if selected and iconName.startswith('folder'):
            iconName = getImageOpen(iconName)
        return iconName

    @patterns.eventSource
    def recomputeAppearance(self, recursive=False, event=None):
        self.__status = None
        # Need to prepare for AttributeError because the cached recursive values
        # are not set in __init__ for performance reasons
        try:
            previousForegroundColor = self.__recursiveForegroundColor
            previousBackgroundColor = self.__recursiveBackgroundColor
            previousRecursiveIcon = self.__recursiveIcon
            previousRecursiveSelectedIcon = self.__recursiveSelectedIcon
        except AttributeError:
            previousForegroundColor = None
            previousBackgroundColor = None
            previousRecursiveIcon = None
            previousRecursiveSelectedIcon = None
        self.__computeRecursiveForegroundColor()
        self.__computeRecursiveBackgroundColor()
        self.__computeRecursiveIcon()
        self.__computeRecursiveSelectedIcon()
        if self.__recursiveForegroundColor != previousForegroundColor or \
           self.__recursiveBackgroundColor != previousBackgroundColor or \
           self.__recursiveIcon != previousRecursiveIcon or \
           self.__recursiveSelectedIcon != previousRecursiveSelectedIcon:
            event.addSource(self, type=self.appearanceChangedEventType())
        if recursive:
            for child in self.children():
                child.recomputeAppearance(recursive=True, event=event)
                
    # percentage Complete
    
    def percentageComplete(self, recursive=False):
        if recursive:
            if self.shouldMarkCompletedWhenAllChildrenCompleted() is None:
                # pylint: disable=E1101    
                ignore_me = self.settings.getboolean('behavior', 
                                'markparentcompletedwhenallchildrencompleted')
            else:
                ignore_me = self.shouldMarkCompletedWhenAllChildrenCompleted()
            percentages = []
            if self.__percentageComplete > 0 or not ignore_me:
                percentages.append(self.__percentageComplete)
            percentages.extend([child.percentageComplete(recursive) for child in self.children()])
            return sum(percentages) / len(percentages) if percentages else 0
        else:
            return self.__percentageComplete
    
    def setPercentageComplete(self, percentage):
        if percentage == self.__percentageComplete:
            return
        oldPercentage = self.__percentageComplete
        self.__percentageComplete = percentage
        if percentage == 100 and oldPercentage != 100 and self.completionDateTime() == self.maxDateTime:
            self.setCompletionDateTime(date.Now())
        elif oldPercentage == 100 and percentage != 100 and self.completionDateTime() != self.maxDateTime:
            self.setCompletionDateTime(self.maxDateTime)
        if 0 < percentage < 100 and self.actualStartDateTime() == date.DateTime():
            self.setActualStartDateTime(date.Now())
        pub.sendMessage(self.percentageCompleteChangedEventType(),
                        newValue=percentage, sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.percentageCompleteChangedEventType(),
                            newValue=ancestor.percentageComplete(recursive=True),
                            sender=ancestor)
    
    @staticmethod
    def percentageCompleteSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.percentageComplete(recursive=recursive)

    @classmethod
    def percentageCompleteSortEventTypes(class_):
        ''' The event types that influence the percentage complete sort order. '''
        return (class_.percentageCompleteChangedEventType(),)

    @classmethod
    def percentageCompleteChangedEventType(class_):
        return 'pubsub.task.percentageComplete'
       
    # priority
    
    def priority(self, recursive=False):
        if recursive:
            childPriorities = [child.priority(recursive=True) \
                               for child in self.children() \
                               if not child.completed()]
            return max(childPriorities + [self.__priority])
        else:
            return self.__priority
        
    def setPriority(self, priority):
        if priority == self.__priority:
            return
        self.__priority = priority
        self.sendPriorityChangedMessage()
    
    def sendPriorityChangedMessage(self):
        pub.sendMessage(self.priorityChangedEventType(), 
                        newValue=self.priority(), sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.priorityChangedEventType(), 
                            newValue=ancestor.priority(),
                            sender=ancestor)

    @classmethod
    def priorityChangedEventType(class_):
        return 'pubsub.task.priority'
    
    @staticmethod
    def prioritySortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.priority(recursive=recursive)

    @classmethod
    def prioritySortEventTypes(class_):
        ''' The event types that influence the priority sort order. '''
        return (class_.priorityChangedEventType(),)
    
    # Hourly fee
    
    def hourlyFee(self, recursive=False):  # pylint: disable=W0613
        return self.__hourlyFee
    
    def setHourlyFee(self, hourlyFee):
        if hourlyFee == self.__hourlyFee:
            return
        self.__hourlyFee = hourlyFee
        pub.sendMessage(self.hourlyFeeChangedEventType(), newValue=hourlyFee, 
                        sender=self)
        if self.timeSpent() > date.TimeDelta():
            self.sendRevenueChangedMessage()
            for effort in self.efforts():
                effort.sendRevenueChangedMessage()
            
    @classmethod
    def hourlyFeeChangedEventType(class_):
        return 'pubsub.task.hourlyFee'
    
    @staticmethod  # pylint: disable=W0613
    def hourlyFeeSortFunction(**kwargs): 
        return lambda task: task.hourlyFee()

    @classmethod
    def hourlyFeeSortEventTypes(class_):
        ''' The event types that influence the hourly fee sort order. '''
        return (class_.hourlyFeeChangedEventType(),)
    
    # Fixed fee
                 
    def fixedFee(self, recursive=False):
        childFixedFees = sum(child.fixedFee(recursive) for child in 
                             self.children()) if recursive else 0
        return self.__fixedFee + childFixedFees
    
    def setFixedFee(self, fixedFee):
        if fixedFee == self.__fixedFee:
            return
        self.__fixedFee = fixedFee
        pub.sendMessage(self.fixedFeeChangedEventType(), newValue=fixedFee,
                        sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.fixedFeeChangedEventType(), 
                            newValue=ancestor.fixedFee(recursive=True),
                            sender=ancestor)
        self.sendRevenueChangedMessage()
        
    @classmethod
    def fixedFeeChangedEventType(class_):
        return 'pubsub.task.fixedFee'

    @staticmethod
    def fixedFeeSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.fixedFee(recursive=recursive)

    @classmethod
    def fixedFeeSortEventTypes(class_):
        ''' The event types that influence the fixed fee sort order. '''
        return (class_.fixedFeeChangedEventType(),)

    # Revenue        
        
    def revenue(self, recursive=False):
        childRevenues = sum(child.revenue(recursive) for child in 
                            self.children()) if recursive else 0
        return self.timeSpent().hours() * self.hourlyFee() + self.fixedFee() + \
               childRevenues

    def sendRevenueChangedMessage(self):
        pub.sendMessage(self.revenueChangedEventType(), 
                        newValue=self.revenue(), sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.revenueChangedEventType(),
                            newValue=ancestor.revenue(recursive=True),
                            sender=ancestor)

    @classmethod
    def revenueChangedEventType(class_):
        return 'pubsub.task.revenue'
    
    @staticmethod
    def revenueSortFunction(**kwargs):            
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.revenue(recursive=recursive)

    @classmethod
    def revenueSortEventTypes(class_):
        ''' The event types that influence the revenue sort order. '''
        return (class_.revenueChangedEventType(),)
    
    # reminder
    
    def reminder(self, recursive=False, includeSnooze=True):  # pylint: disable=W0613
        if recursive:
            reminders = [child.reminder(recursive=True) for child in \
                         self.children()] + [self.__reminder]
            reminders = [reminder for reminder in reminders if reminder]
            return min(reminders) if reminders else None
        else:
            return self.__reminder if includeSnooze else self.__reminderBeforeSnooze

    def setReminder(self, reminderDateTime=None):
        if reminderDateTime == self.maxDateTime:
            reminderDateTime = None
        if reminderDateTime == self.__reminder:
            return
        self.__reminder = reminderDateTime
        self.__reminderBeforeSnooze = reminderDateTime
        pub.sendMessage(self.reminderChangedEventType(), 
                        newValue=reminderDateTime, sender=self)
        for ancestor in self.ancestors():
            pub.sendMessage(ancestor.reminderChangedEventType(),
                            newValue=reminderDateTime, sender=ancestor)
        
    def snoozeReminder(self, timeDelta, now=date.Now):
        if timeDelta:
            self.__reminder = now() + timeDelta
            pub.sendMessage(self.reminderChangedEventType(), 
                            newValue=self.__reminder, sender=self)
        else:
            if self.recurrence():
                self.__reminder = None
                pub.sendMessage(self.reminderChangedEventType(), 
                                newValue=self.__reminder, sender=self)
            else:
                self.setReminder()

    @classmethod
    def reminderChangedEventType(class_):
        return 'pubsub.task.reminder'
    
    @staticmethod
    def reminderSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        maxDateTime = date.DateTime()
        return lambda task: task.reminder(recursive=recursive) or maxDateTime

    @classmethod
    def reminderSortEventTypes(class_):
        ''' The event types that influence the reminder sort order. '''
        return (class_.reminderChangedEventType(),)

    # Recurrence
    
    def recurrence(self, recursive=False, upwards=False):
        if not self.__recurrence and recursive and upwards and self.parent():
            return self.parent().recurrence(recursive, upwards)
        elif recursive and not upwards:
            recurrences = [child.recurrence() for child in self.children(recursive)]
            recurrences.append(self.__recurrence)
            recurrences = [r for r in recurrences if r]
            return min(recurrences) if recurrences else self.__recurrence
        else:
            return self.__recurrence

    def setRecurrence(self, recurrence=None):
        recurrence = recurrence or date.Recurrence()
        if recurrence == self.__recurrence:
            return
        self.__recurrence = recurrence
        pub.sendMessage(self.recurrenceChangedEventType(), newValue=recurrence,
                        sender=self)
        
    @classmethod
    def recurrenceChangedEventType(class_):
        return 'pubsub.task.recurrence'

    @patterns.eventSource
    def recur(self, completionDateTime=None, event=None):
        completionDateTime = completionDateTime or date.Now()
        self.setCompletionDateTime(self.maxDateTime)
        recur = self.recurrence(recursive=True, upwards=True)
        
        currentDueDateTime = self.dueDateTime()
        if currentDueDateTime != date.DateTime():        
            basisForRecurrence = completionDateTime if recur.recurBasedOnCompletion else currentDueDateTime 
            nextDueDateTime = recur(basisForRecurrence, next=False)
            nextDueDateTime = nextDueDateTime.replace(hour=currentDueDateTime.hour,
                                                      minute=currentDueDateTime.minute,
                                                      second=currentDueDateTime.second,
                                                      microsecond=currentDueDateTime.microsecond)
            self.setDueDateTime(nextDueDateTime)
        
        currentPlannedStartDateTime = self.plannedStartDateTime()
        if currentPlannedStartDateTime != date.DateTime():        
            if date.DateTime() not in (currentPlannedStartDateTime, currentDueDateTime):
                taskDuration = currentDueDateTime - currentPlannedStartDateTime
                nextPlannedStartDateTime = nextDueDateTime - taskDuration
            else:
                basisForRecurrence = completionDateTime if recur.recurBasedOnCompletion else currentPlannedStartDateTime
                nextPlannedStartDateTime = recur(basisForRecurrence, next=False)
            nextPlannedStartDateTime = nextPlannedStartDateTime.replace(hour=currentPlannedStartDateTime.hour,
                                                                        minute=currentPlannedStartDateTime.minute,
                                                                        second=currentPlannedStartDateTime.second,
                                                                        microsecond=currentPlannedStartDateTime.microsecond)
            self.setPlannedStartDateTime(nextPlannedStartDateTime)
        
        self.setActualStartDateTime(date.DateTime())
        self.setPercentageComplete(0)
        if self.reminder(includeSnooze=False):
            nextReminder = recur(self.reminder(includeSnooze=False), next=False)
            self.setReminder(nextReminder)
        for child in self.children():
            if not child.recurrence():
                child.recur(completionDateTime, event=event)
        self.recurrence()(next=True)

    @staticmethod
    def recurrenceSortFunction(**kwargs):
        recursive = kwargs.get('treeMode', False)
        return lambda task: task.recurrence(recursive=recursive)

    @classmethod
    def recurrenceSortEventTypes(class_):
        ''' The event types that influence the recurrence sort order. '''
        return (class_.recurrenceChangedEventType(),)

    # Prerequisites
    
    def prerequisites(self, recursive=False, upwards=False):
        prerequisites = set(self.__prerequisites)
        if recursive and upwards and self.parent() is not None:
            prerequisites |= self.parent().prerequisites(recursive=True, upwards=True)
        elif recursive and not upwards:
            for child in self.children(recursive=True):
                prerequisites |= child.prerequisites()
        return prerequisites
    
    def setPrerequisites(self, prerequisites):
        prerequisites = set(prerequisites)
        if prerequisites == self.prerequisites():
            return
        self.__prerequisites = WeakSet(prerequisites)
        self.setActualStartDateTime(self.maxDateTime, recursive=True)
        self.recomputeAppearance(recursive=True)
        pub.sendMessage(self.prerequisitesChangedEventType(), 
                        newValue=self.prerequisites(), sender=self)
  
    def addPrerequisites(self, prerequisites):
        prerequisites = set(prerequisites)
        if prerequisites <= self.prerequisites():
            return
        self.__prerequisites = WeakSet(prerequisites | self.prerequisites())
        self.setActualStartDateTime(self.maxDateTime, recursive=True)
        self.recomputeAppearance(recursive=True)
        pub.sendMessage(self.prerequisitesChangedEventType(), 
                        newValue=self.prerequisites(), sender=self)
        
    def removePrerequisites(self, prerequisites):
        prerequisites = set(prerequisites)
        if self.prerequisites().isdisjoint(prerequisites):
            return
        self.__prerequisites = WeakSet(self.prerequisites() - prerequisites)
        self.recomputeAppearance(recursive=True)
        pub.sendMessage(self.prerequisitesChangedEventType(), 
                        newValue=self.prerequisites(), sender=self)
        
    def addTaskAsDependencyOf(self, prerequisites):
        for prerequisite in prerequisites:
            prerequisite.addDependencies([self])
    
    def removeTaskAsDependencyOf(self, prerequisites):
        for prerequisite in prerequisites:
            prerequisite.removeDependencies([self])
            
    @classmethod
    def prerequisitesChangedEventType(class_):
        return 'pubsub.task.prerequisites'

    @staticmethod
    def prerequisitesSortFunction(**kwargs):
        ''' Return a sort key for sorting by prerequisites. Since a task can 
            have multiple prerequisites we first sort the prerequisites by their
            subjects. If the sorter is in tree mode, we also take the 
            prerequisites of the children of the task into account, after the 
            prerequisites of the task itself. If the sorter is in list
            mode we also take the prerequisites of the parent (recursively) into
            account, again after the prerequisites of the categorizable itself.'''
        def sortKeyFunction(task):
            def sortedSubjects(items):
                return sorted([item.subject(recursive=True) for item in items])
            prerequisites = task.prerequisites()
            sortedPrerequisiteSubjects = sortedSubjects(prerequisites)
            isListMode = not kwargs.get('treeMode', False)
            childPrerequisites = task.prerequisites(recursive=True, upwards=isListMode) - prerequisites
            sortedPrerequisiteSubjects.extend(sortedSubjects(childPrerequisites)) 
            return sortedPrerequisiteSubjects
        return sortKeyFunction

    @classmethod
    def prerequisitesSortEventTypes(class_):
        ''' The event types that influence the prerequisites sort order. '''
        return (class_.prerequisitesChangedEventType(),)

    # Dependencies
    
    def dependencies(self, recursive=False, upwards=False):
        dependencies = set(self.__dependencies)
        if recursive and upwards and self.parent() is not None:
            dependencies |= self.parent().dependencies(recursive=True, upwards=True)
        elif recursive and not upwards:
            for child in self.children(recursive=True):
                dependencies |= child.dependencies()
        return dependencies

    def setDependencies(self, dependencies):
        dependencies = set(dependencies)
        if dependencies == self.dependencies():
            return
        self.__dependencies = WeakSet(dependencies)
        pub.sendMessage(self.dependenciesChangedEventType(),
                        newValue=self.dependencies(), sender=self)
    
    def addDependencies(self, dependencies):
        dependencies = set(dependencies)
        if dependencies <= self.dependencies():
            return
        self.__dependencies = WeakSet(self.dependencies() | dependencies)
        pub.sendMessage(self.dependenciesChangedEventType(),
                        newValue=self.dependencies(), sender=self)

    def removeDependencies(self, dependencies):
        dependencies = set(dependencies)
        if self.dependencies().isdisjoint(dependencies):
            return
        self.__dependencies = WeakSet(self.dependencies() - dependencies)
        pub.sendMessage(self.dependenciesChangedEventType(),
                        newValue=self.dependencies(), sender=self)
        
    def addTaskAsPrerequisiteOf(self, dependencies):
        for dependency in dependencies:
            dependency.addPrerequisites([self])
            
    def removeTaskAsPrerequisiteOf(self, dependencies):
        for dependency in dependencies:
            dependency.removePrerequisites([self])      

    @classmethod
    def dependenciesChangedEventType(class_):
        return 'pubsub.task.dependencies'
    
    @staticmethod
    def dependenciesSortFunction(**kwargs):
        ''' Return a sort key for sorting by dependencies. Since a task can 
            have multiple dependencies we first sort the dependencies by their
            subjects. If the sorter is in tree mode, we also take the 
            dependencies of the children of the task into account, after the 
            dependencies of the task itself. If the sorter is in list
            mode we also take the dependencies of the parent (recursively) into
            account, again after the dependencies of the categorizable itself.'''
        def sortKeyFunction(task):
            def sortedSubjects(items):
                return sorted([item.subject(recursive=True) for item in items])
            dependencies = task.dependencies()
            sortedDependencySubjects = sortedSubjects(dependencies)
            isListMode = not kwargs.get('treeMode', False)
            childDependencies = task.dependencies(recursive=True, upwards=isListMode) - dependencies
            sortedDependencySubjects.extend(sortedSubjects(childDependencies)) 
            return sortedDependencySubjects
        return sortKeyFunction

    @classmethod
    def dependenciesSortEventTypes(class_):
        ''' The event types that influence the dependencies sort order. '''
        return (class_.dependenciesChangedEventType(),)
                
    # behavior
    
    def setShouldMarkCompletedWhenAllChildrenCompleted(self, newValue):
        if newValue == self.__shouldMarkCompletedWhenAllChildrenCompleted:
            return
        self.__shouldMarkCompletedWhenAllChildrenCompleted = newValue
        pub.sendMessage(self.shouldMarkCompletedWhenAllChildrenCompletedChangedEventType(),
                        newValue=newValue, sender=self)
        pub.sendMessage(self.percentageCompleteChangedEventType(), 
                        newValue=self.percentageComplete(), sender=self)
    
    @classmethod
    def shouldMarkCompletedWhenAllChildrenCompletedChangedEventType(class_):
        return 'pubsub.task.shouldMarkCompletedWhenAllChildrenCompleted'

    def shouldMarkCompletedWhenAllChildrenCompleted(self):
        return self.__shouldMarkCompletedWhenAllChildrenCompleted

    @classmethod
    def suggestedPlannedStartDateTime(cls, now=date.Now):
        return cls.suggestedDateTime('defaultplannedstartdatetime', now)

    @classmethod
    def suggestedActualStartDateTime(cls, now=date.Now):
        return cls.suggestedDateTime('defaultactualstartdatetime', now)
     
    @classmethod
    def suggestedDueDateTime(cls, now=date.Now):
        return cls.suggestedDateTime('defaultduedatetime', now)
        
    @classmethod
    def suggestedCompletionDateTime(cls, now=date.Now):
        return cls.suggestedDateTime('defaultcompletiondatetime', now)
    
    @classmethod
    def suggestedReminderDateTime(cls, now=date.Now):
        return cls.suggestedDateTime('defaultreminderdatetime', now)
    
    @classmethod    
    def suggestedDateTime(cls, defaultDateTimeSetting, now=date.Now):
        # pylint: disable=E1101,W0142
        defaultDateTime = cls.settings.get('view', defaultDateTimeSetting)
        dummy_prefix, defaultDate, defaultTime = defaultDateTime.split('_')
        dateTime = now()
        currentTime = dict(hour=dateTime.hour, minute=dateTime.minute,
                           second=dateTime.second, microsecond=dateTime.microsecond)
        if defaultDate == 'tomorrow':
            dateTime += date.ONE_DAY
        elif defaultDate == 'dayaftertomorrow':
            dateTime += (date.ONE_DAY + date.ONE_DAY)
        elif defaultDate == 'nextfriday':
            dateTime = (dateTime + date.ONE_DAY).endOfWorkWeek().replace(**currentTime)
        elif defaultDate == 'nextmonday':
            dateTime = (dateTime + date.ONE_WEEK).startOfWorkWeek().replace(**currentTime)
            
        if defaultTime == 'startofday':
            return dateTime.startOfDay()
        elif defaultTime == 'startofworkingday':
            startHour = cls.settings.getint('view', 'efforthourstart')
            return dateTime.replace(hour=startHour, minute=0,
                                    second=0, microsecond=0)
        elif defaultTime == 'currenttime':
            return dateTime
        elif defaultTime == 'endofworkingday':
            endHour = cls.settings.getint('view', 'efforthourend')
            if endHour >= 24:
                endHour, minute, second = 23, 59, 59
            else:
                minute, second = 0, 0
            return dateTime.replace(hour=endHour, minute=minute,
                                    second=second, microsecond=0)
        elif defaultTime == 'endofday':
            return dateTime.endOfDay()
        
    @classmethod
    def modificationEventTypes(class_):
        eventTypes = super(Task, class_).modificationEventTypes()
        return eventTypes + [class_.plannedStartDateTimeChangedEventType(), 
                             class_.dueDateTimeChangedEventType(), 
                             class_.actualStartDateTimeChangedEventType(),
                             class_.completionDateTimeChangedEventType(),
                             class_.effortsChangedEventType(),
                             class_.budgetChangedEventType(),
                             class_.percentageCompleteChangedEventType(),
                             class_.priorityChangedEventType(),
                             class_.hourlyFeeChangedEventType(), 
                             class_.fixedFeeChangedEventType(),
                             class_.reminderChangedEventType(), 
                             class_.recurrenceChangedEventType(),
                             class_.prerequisitesChangedEventType(),
                             class_.dependenciesChangedEventType(),
                             class_.shouldMarkCompletedWhenAllChildrenCompletedChangedEventType()]



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.i18n import _
from taskcoachlib.domain import category
import base


class NewCategoryCommand(base.NewItemCommand):
    singular_name = _('New category')
    
    def __init__(self, *args, **kwargs):
        subject = kwargs.pop('subject', _('New category'))
        description = kwargs.pop('description', '')
        attachments = kwargs.pop('attachments', [])
        super(NewCategoryCommand, self).__init__(*args, **kwargs)
        self.items = self.createNewCategories(subject=subject, 
            description=description, attachments=attachments)

    def createNewCategories(self, **kwargs):
        return [category.Category(**kwargs)]
        

class NewSubCategoryCommand(base.NewSubItemCommand):
    plural_name = _('New subcategories')
    singular_name = _('New subcategory of "%s"')

    def __init__(self, *args, **kwargs):
        subject = kwargs.pop('subject', _('New subcategory'))
        description = kwargs.pop('description', '')
        attachments = kwargs.pop('attachments', [])
        super(NewSubCategoryCommand, self).__init__(*args, **kwargs)
        self.items = self.createNewCategories(subject=subject,
            description=description, attachments=attachments)
        self.save_modification_datetimes()

    def createNewCategories(self, **kwargs):
        return [parent.newChild(**kwargs) for parent in self.items]
    

class EditExclusiveSubcategoriesCommand(base.BaseCommand):
    plural_name = _('Edit exclusive subcategories')
    singular_name = _('Edit exclusive subcategories of "%s"')

    def __init__(self, *args, **kwargs):
        self.__newExclusivity = kwargs.pop('newValue')
        super(EditExclusiveSubcategoriesCommand, self).__init__(*args, **kwargs)
        self.__oldExclusivities = [item.hasExclusiveSubcategories() for item in self.items]
        
    @patterns.eventSource
    def do_command(self, event=None):
        super(EditExclusiveSubcategoriesCommand, self).do_command()
        for item in self.items:
            item.makeSubcategoriesExclusive(self.__newExclusivity, event=event)

    @patterns.eventSource
    def undo_command(self, event=None):
        super(EditExclusiveSubcategoriesCommand, self).undo_command()
        for item, oldExclusivity in zip(self.items, self.__oldExclusivities):
            item.makeSubcategoriesExclusive(oldExclusivity, event=event)

    def redo_command(self):
        self.do_command()


class DeleteCategoryCommand(base.DeleteCommand):
    plural_name = _('Delete categories')
    singular_name = _('Delete category "%s"')
    
    
class DragAndDropCategoryCommand(base.OrderingDragAndDropCommand):
    plural_name = _('Drag and drop categories')



# https://github.com/kfdm/gntp/issues/27
# nosetests -v gntp.test.test_resources:ResourceTest

import os
from gntp.test import GNTPTestCase
import gntp

ICON_FILE = os.path.join(os.path.dirname(__file__), "growl-icon.png")
ICON_DATA = open(ICON_FILE, 'rb').read()
FILE_DATA = open(__file__).read()


class ResourceTest(GNTPTestCase):
	def test_single_resource(self):
		notification = gntp.GNTPNotice(
			app=self.application,
			name=self.notification_name,
			title="Testing Single Resource",
			password=self.growl.password,
			)
		resource = notification.add_resource(ICON_DATA)
		notification.add_header('Notification-Icon', resource)
		self.assertIsTrue(self.growl._send('notify', notification))

	def test_double_resource(self):
		notification = gntp.GNTPNotice(
			app=self.application,
			name=self.notification_name,
			title="Testing Double Resource",
			password=self.growl.password,
			)

		notification.add_resource(FILE_DATA)
		resource = notification.add_resource(ICON_DATA)
		notification.add_header('Notification-Icon', resource)

		self.assertIsTrue(self.growl._send('notify', notification))



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from notifier import AbstractNotifier
import tempfile, os, wx
from taskcoachlib.thirdparty import snarl


class SnarlNotifier(AbstractNotifier):
    def getName(self):
        return u'Snarl'

    def isAvailable(self):
        try:
            return bool(snarl.snGetVersion())
        except:
            return False

    def notify(self, title, summary, bitmap, **kwargs):
        # Hum. Snarl needs a file.
        fd, filename = tempfile.mkstemp('.png')
        os.close(fd)
        bitmap.SaveFile(filename, wx.BITMAP_TYPE_PNG)
        try:
            snarl.snShowMessage(title.encode('UTF-8'), summary.encode('UTF-8'), iconPath=filename)
        finally:
            os.remove(filename)


AbstractNotifier.register(SnarlNotifier())



#Python Snarl bindings version 0.2a
#Hacked together by Alexander Lash in the wee hours of 2/27.
#Contact him at alexander.lash@gmail.com
#0.1a: Initial release
#0.2a: Added basic command line functions
#LARGELY UNTESTED REFERENCE IMPLEMENTATION
import win32gui, win32api, win32con
import struct, array
from ctypes import cast, POINTER, c_byte

class SNARL_COMMANDS:
    SNARL_SHOW = 1
    SNARL_HIDE = 2
    SNARL_UPDATE = 3
    SNARL_IS_VISIBLE = 4
    SNARL_GET_VERSION = 5
    SNARL_REGISTER_CONFIG_WINDOW = 6
    SNARL_REVOKE_CONFIG_WINDOW = 7
    SNARL_REGISTER_ALERT = 8
    SNARL_REVOKE_ALERT = 9
    SNARL_REGISTER_CONFIG_WINDOW_2 = 0xA
    SNARL_EX_SHOW = 0x20
    
    @staticmethod
    def sendCommand(command, id=0, timeout=0, longdata=0, title="", text="", icon="",
                      extra=None, extra2=None, reserved1=None, reserved2=None):
        if extra is None and extra2 is None and reserved1 is None and reserved2 is None:
            command = struct.pack("ILLL1024s1024s1024s",
                              command,
                              id,
                              timeout,
                              longdata, #LngData2
                              array.array('B', title).tostring(),
                              array.array('B', text).tostring(),
                              array.array('B', icon).tostring())
        else:
            if reserved1 is None:
                reserved1 = 0
            if reserved2 is None:
                reserved2 = 0
            if extra is None:
                extra = ""
            if extra2 is None:
                extra2 = ""
            command = struct.pack("ILLL1024s1024s1024s1024s1024sLL",
                                  command,
                                  id,
                                  timeout,
                                  longdata, #LngData2
                                  array.array('B', title).tostring(),
                                  array.array('B', text).tostring(),
                                  array.array('B', icon).tostring(),
                                  array.array('B', extra).tostring(),
                                  array.array('B', extra2).tostring(),
                                  reserved1,
                                  reserved2)
        command_pack = array.array("B", command)
        command_info = command_pack.buffer_info()
        
        cd = struct.pack("LLP", 2, command_info[1], command_info[0])
        cd_pack = array.array("B", cd)
        cd_info = cd_pack.buffer_info()
        
        
        hwnd = win32gui.FindWindow(None, 'Snarl')
        if hwnd:
            return win32api.SendMessage(hwnd, win32con.WM_COPYDATA, 0, cd_info[0])
        else:
            return False


#deviation from spec: Python doesn't do references.
def snGetVersion():
    hr = SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_GET_VERSION)
    if hr:
        ver = cast(hr, POINTER(c_byte))
        return (ver[0], ver[1])
    return False

#deviation from spec: sound is an optional parameter.
def snShowMessage(title, text, timeout=0, iconPath="", reply=0, reply_msg=0, sound=None):
    if title is None or text is None:
        return False
    if sound is None:
        return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_SHOW, reply, timeout, reply_msg,
                        title,text,iconPath)
    return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_SHOW, reply, timeout, reply_msg,
                        title,text,iconPath,sound)

def snUpdateMessage(id, title, text):
    if id is None or title is None or text is None:
        return False
    return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_UPDATE, id=id, title=title, text=text) == -1

def snHideMessage(id):
    id = int(id)
    if id is None:
        return False
    return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_HIDE, id=id) == -1

#For some reason this ALWAYS returns -1 for all messages that were once displayed.
def snIsMessageVisible(id):
    return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_IS_VISIBLE, id=id) == -1

#Untested
def snRegisterConfig(hwnd, appname, reply_msg, icon=None):
    if icon is None:
        return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_REGISTER_CONFIG_WINDOW,
                            longdata=hwnd, title=appname, id = reply_msg)
    return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_REGISTER_CONFIG_WINDOW,
                            longdata=hwnd, title=appname, id = reply_msg,
                            icon=icon)

#Untested
def snRevokeConfig(hwnd):
    return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_REVOKE_CONFIG_WINDOW,
                        longdata=hwnd)

#Reference implementation - seems to do nothing?
def snRegisterAlert(title, text):
    return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_REGISTER_ALERT,
                        title=title, text=text)

#Reference implementation - seems to do nothing?
def snRevokeAlert():
    return SNARL_COMMANDS.sendCommand(SNARL_COMMANDS.SNARL_REVOKE_ALERT)

if __name__=='__main__':
    import sys, inspect
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option("-s", "--show", action="store_const", dest="cmd",
                      const=snShowMessage,
                      help="Show a message with the given parameters.")
    parser.add_option("-u", "--update", action="store_const", dest="cmd",
                      const=snUpdateMessage,
                      help="Update a message with the given parameters.")
    parser.add_option("-c", "--hide", action="store_const", dest="cmd",
                      const=snHideMessage,
                      help="Hide a message with the given ID.")
    parser.add_option("-i", "--id", dest="id",
                      help="ID to use for the update/hide.")
    parser.add_option("-t", "--title", dest="title",
                      help="Title to use for the show/update.")
    parser.add_option("-x", "--text", dest="text",
                      help="Text to use for the show/update.")
    parser.add_option("-T", "--timeout", dest="timeout",
                      help="Timeout to use for the show.")
    parser.add_option("-I", "--icon", dest="iconPath",
                      help="Icon to use for the show.")
    parser.add_option("-S", "--sound", dest="sound",
                      help="Sound to use for the show.")
    if snGetVersion == False:
        print "Snarl not running!"
        sys.exit(1)
    (options, args) = parser.parse_args(sys.argv[1:])
    if options.cmd is None:
        parser.parse_args(['-h'])
    d = options.__dict__.copy()
    cmd = options.cmd
    for key in options.__dict__:
        if key not in inspect.getargspec(cmd)[0]:
            del d[key]
        elif d[key] is None:
            del d[key]
    print cmd(**d)
    sys.exit(0)


'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''


class Singleton(type):
    ''' Singleton metaclass. Use by defining the metaclass of a class Singleton,
        e.g.: class ThereCanBeOnlyOne:
                  __metaclass__ = Singleton 
    '''              

    def __call__(class_, *args, **kwargs):
        if not class_.hasInstance():
            # pylint: disable=W0201
            class_.instance = super(Singleton, class_).__call__(*args, **kwargs)
        return class_.instance

    def deleteInstance(class_):
        ''' Delete the (only) instance. This method is mainly for unittests so
            they can start with a clean slate. '''
        if class_.hasInstance():
            del class_.instance

    def hasInstance(class_):
        ''' Has the (only) instance been created already? '''
        return 'instance' in class_.__dict__



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, locale
from taskcoachlib import widgets, render
from taskcoachlib.domain import date


class CommonTestsMixin(object):
    def setUp(self):
        super(CommonTestsMixin, self).setUp()
        # LC_ALL does not work on Slackware or Arch, but LC_TIME crashes on Fedora...
        try:
            self.__oldLocale = locale.getlocale(locale.LC_ALL)
            self.__localeDomain = locale.LC_ALL
        except TypeError:
            self.__oldLocale = locale.getlocale(locale.LC_TIME)
            self.__localeDomain = locale.LC_TIME
        localeName = 'en_US' if self.ampm else 'fr_FR'
        # OS X and Linux don't agree on encoding names...
        for encodingName in ['utf8', 'UTF-8']:
            try:
                locale.setlocale(self.__localeDomain, '%s.%s' % (localeName, encodingName))
            except locale.Error:
                pass
            else:
                break
        else:
            self.skipTest('No supported locale found. Try "locale -a" and add a supported locale.')
        reload(render) # To execute module-level code every time

    def tearDown(self):
        locale.setlocale(self.__localeDomain, self.__oldLocale)
        reload(render)
        super(CommonTestsMixin, self).tearDown()

    def _format(self, hour, minute, second):
        if self.ampm:
            dpyHour = hour % 12
            if dpyHour == 0:
                dpyHour = 12
            r = '%02d:%02d' % (dpyHour, minute)
            if self.showSeconds:
                r += ':%02d' % second
            r += ' AM' if hour <= 12 else ' PM'
        else:
            r = '%02d:%02d' % (hour, minute)
            if self.showSeconds:
                r += ':%02d' % second
        return r

    def testGetValue(self):
        oneHour = date.DateTime(2000, 1, 1, hour=1)
        self.dateTimeCtrl.SetValue(oneHour)
        self.assertEqual(oneHour, self.dateTimeCtrl.GetValue())

        
class DateTimeCtrlTestCase(test.wxTestCase):
    adjustEndOfDay = False
    showSeconds = False

    def setUp(self):
        super(DateTimeCtrlTestCase, self).setUp()
        self.dateTimeCtrl = widgets.datectrl.DateTimeCtrl(self.frame, 
                                                          showSeconds=self.showSeconds,
                                                          adjustEndOfDay=self.adjustEndOfDay)

    def test_adjust(self):
        self.dateTimeCtrl.SetValue(date.DateTime(2012, 12, 12, 23, 59, 0, 0))
        self.assertEqual(self.dateTimeCtrl.GetValue(),
                         date.DateTime(2012, 12, 12, 23, 59, 59, 999999) if self.adjustEndOfDay \
                             else date.DateTime(2012, 12, 12, 23, 59, 0, 0))


class DateTimeCtrlTest_Seconds_Base(CommonTestsMixin):
    showSeconds = True

    def testGetValue_SecondPrecision(self):
        oneHourAndTenSeconds = date.DateTime(2000, 1, 1, hour=1, second=10)
        self.dateTimeCtrl.SetValue(oneHourAndTenSeconds)
        self.assertEqual(oneHourAndTenSeconds, self.dateTimeCtrl.GetValue())


class DateTimeCtrlTest_Seconds(DateTimeCtrlTest_Seconds_Base, DateTimeCtrlTestCase):
    ampm = False


class DateTimeCtrlTest_Seconds_AMPM(DateTimeCtrlTest_Seconds_Base, DateTimeCtrlTestCase):
    ampm = True


class DateTimeCtrlTest_NoSeconds_Base(CommonTestsMixin):
    showSeconds = False

    def testGetValue_SecondPrecision(self):
        oneHour = date.DateTime(2000, 1, 1, hour=1)
        oneHourAndTenSeconds = date.DateTime(2000, 1, 1, hour=1, second=10)
        self.dateTimeCtrl.SetValue(oneHourAndTenSeconds)
        self.assertEqual(oneHour, self.dateTimeCtrl.GetValue())


class DateTimeCtrlTest_NoSeconds(DateTimeCtrlTest_NoSeconds_Base, DateTimeCtrlTestCase):
    ampm = False


class DateTimeCtrlTest_NoSeconds_Adjust(DateTimeCtrlTest_NoSeconds_Base, DateTimeCtrlTestCase):
    ampm = False
    adjustEndOfDay = True


class DateTimeCtrlTest_NoSeconds_AMPM(DateTimeCtrlTest_NoSeconds_Base, DateTimeCtrlTestCase):
    ampm = True


class DateTimeCtrlTest_NoSeconds_AMPM_Adjust(DateTimeCtrlTest_NoSeconds_Base, DateTimeCtrlTestCase):
    ampm = True
    adjustEndOfDay = True



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from unittests import asserts 
from taskcoachlib import patterns


class HistoryTest(test.TestCase, asserts.CommandAssertsMixin):
    def setUp(self):
        self.commands = patterns.CommandHistory()
        self.command = patterns.Command()

    def tearDown(self):
        self.commands.clear()

    def testSingleton(self):
        another = patterns.CommandHistory()
        self.failUnless(self.commands is another)

    def testClear(self):
        self.command.do()
        self.commands.clear()
        self.assertHistoryAndFuture([], [])

    def testDo(self):
        self.command.do()
        self.assertHistoryAndFuture([self.command], [])

    def testUndo(self):
        self.command.do()
        self.commands.undo()
        self.assertHistoryAndFuture([], [self.command])

    def testRedo(self):
        self.command.do()
        self.commands.undo()
        self.commands.redo()
        self.assertHistoryAndFuture([self.command], [])

    def testUndoStr_EmptyHistory(self):
        self.assertEqual('Undo', self.commands.undostr())

    def testUndoStr(self):
        self.command.do()
        self.assertEqual('Undo %s'%self.command, self.commands.undostr())

    def testRedoStr_EmptyFuture(self):
        self.assertEqual('Redo', self.commands.redostr())

    def testRedoStr(self):
        self.command.do()
        self.commands.undo()
        self.assertEqual('Redo %s'%self.command, self.commands.redostr())

    def testHasHistory(self):
        self.failIf(self.commands.hasHistory())
        self.command.do()
        self.failUnless(self.commands.hasHistory())
        self.commands.undo()
        self.failIf(self.commands.hasHistory())

    def testHasFuture(self):
        self.command.do()
        self.failIf(self.commands.hasFuture())
        self.commands.undo()
        self.failUnless(self.commands.hasFuture())
        self.commands.redo()
        self.failIf(self.commands.hasFuture())

